import time

from ..request import Request
from .Provider import Provider

from ..configuration import config
from ..presets.PresetsCapsule import PresetsCapsule
from ..presets import Tailwind, Vue, React, Bootstrap
from ..cors import Cors


class FrameworkProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        # @M5 remove this and add PresetsProvider in default project
        presets = PresetsCapsule()
        presets.add(Bootstrap())
        presets.add(Tailwind())
        presets.add(Vue())
        presets.add(React())
        self.application.bind("presets", presets)

        # @M5 remove this and add SecurityProvider in default project
        # @M5 old projects won't have securiy options so put default here. remove this for M5.
        options = config("security.cors")
        if not options:
            options = {
                "paths": ["api/*"],
                "allowed_methods": ["*"],
                "allowed_origins": ["*"],
                "allowed_headers": ["*"],
                "exposed_headers": [],
                "max_age": None,
                "supports_credentials": False,
            }
        cors = Cors(self.application).set_options(options)
        self.application.bind("cors", cors)

    def boot(self):
        from ..response import Response

        request = Request(self.application.make("environ"))
        request.app = self.application
        if self.application.has("activate.subdomains") and self.application.make(
            "activate.subdomains"
        ):
            request.activate_subdomains()
        self.application.bind("request", request)
        self.application.bind("response", Response(self.application))

        self.application.bind("start_time", time.time())



from masonite.middleware import Middleware

from masonite.api.facades import Api


class JWTAuthenticationMiddleware(Middleware):
    def before(self, request, response):
        token = Api.get_token()
        if not token:
            return response.json(
                {"message": "Authentication token missing"}, status="401"
            )

        # Check token is not expired
        validate = Api.validate_token(token)
        if not validate:
            return response.json(
                {"message": "Token invalid. Try reauthenticating."}, status=401
            )

        return request

    def after(self, request, response):
        pass



import json
import random
import operator
from functools import reduce
from dotty_dict import Dotty

from .structures import data_get


class Collection:
    """Wraps various data types to make working with them easier."""

    def __init__(self, items=None):
        self._items = items or []
        self.__appends__ = []

    def take(self, number: int):
        """Takes a specific number of results from the items.

        Arguments:
            number {integer} -- The number of results to take.

        Returns:
            int
        """
        if number < 0:
            return self[number:]

        return self[:number]

    def first(self, callback=None):
        """Takes the first result in the items.

        If a callback is given then the first result will be the result after the filter.

        Keyword Arguments:
            callback {callable} -- Used to filter the results before returning the first item. (default: {None})

        Returns:
            mixed -- Returns whatever the first item is.
        """
        filtered = self
        if callback:
            filtered = self.filter(callback)
        response = None
        if filtered:
            response = filtered[0]
        return response

    def last(self, callback=None):
        """Takes the last result in the items.

        If a callback is given then the last result will be the result after the filter.

        Keyword Arguments:
            callback {callable} -- Used to filter the results before returning the last item. (default: {None})

        Returns:
            mixed -- Returns whatever the last item is.
        """
        filtered = self
        if callback:
            filtered = self.filter(callback)
        return filtered[-1]

    def all(self):
        """Returns all the items.

        Returns:
            mixed -- Returns all items.
        """
        return self._items

    def avg(self, key=None):
        """Returns the average of the items.

        If a key is given it will return the average of all the values of the key.

        Keyword Arguments:
            key {string} -- The key to use to find the average of all the values of that key. (default: {None})

        Returns:
            int -- Returns the average.
        """
        result = 0
        items = self._get_value(key) or self._items
        try:
            result = sum(items) / len(items)
        except TypeError:
            pass
        return result

    def max(self, key=None):
        """Returns the average of the items.

        If a key is given it will return the average of all the values of the key.

        Keyword Arguments:
            key {string} -- The key to use to find the average of all the values of that key. (default: {None})

        Returns:
            int -- Returns the average.
        """
        result = 0
        items = self._get_value(key) or self._items

        try:
            return max(items)
        except (TypeError, ValueError):
            pass
        return result

    def chunk(self, size: int):
        """Chunks the items.

        Keyword Arguments:
            size {integer} -- The number of values in each chunk.

        Returns:
            int -- Returns the average.
        """
        items = []
        for i in range(0, self.count(), size):
            items.append(self[i : i + size])
        return self.__class__(items)

    def collapse(self):
        items = []
        for item in self:
            items += self.__get_items(item)
        return self.__class__(items)

    def contains(self, key, value=None):
        if value:
            return self.contains(lambda x: self._data_get(x, key) == value)

        if self._check_is_callable(key, raise_exception=False):
            return self.first(key) is not None

        return key in self

    def count(self):
        return len(self._items)

    def diff(self, items):
        items = self.__get_items(items)
        return self.__class__([x for x in self if x not in items])

    def each(self, callback):
        self._check_is_callable(callback)

        for k, v in enumerate(self):
            result = callback(v)
            if not result:
                break
            self[k] = result

    def every(self, callback):
        self._check_is_callable(callback)
        return all([callback(x) for x in self])

    def filter(self, callback):
        self._check_is_callable(callback)
        return self.__class__(list(filter(callback, self)))

    def flatten(self):
        def _flatten(items):
            if isinstance(items, dict):
                for v in items.values():
                    for x in _flatten(v):
                        yield x
            elif isinstance(items, list):
                for i in items:
                    for j in _flatten(i):
                        yield j
            else:
                yield items

        return self.__class__(list(_flatten(self._items)))

    def forget(self, *keys):
        keys = reversed(sorted(keys))

        for key in keys:
            del self[key]

        return self

    def for_page(self, page, number):
        return self.__class__(self[page:number])

    def get(self, key, default=None):
        try:
            return self[key]
        except IndexError:
            pass

        return self._value(default)

    def implode(self, glue=",", key=None):
        first = self.first()
        if not isinstance(first, str) and key:
            return glue.join(self.pluck(key))
        return glue.join([str(x) for x in self])

    def is_empty(self):
        return not self

    def map(self, callback):
        self._check_is_callable(callback)
        items = [callback(x) for x in self]
        return self.__class__(items)

    def map_into(self, cls, method=None, **kwargs):
        results = []
        for item in self:
            if method:
                results.append(getattr(cls, method)(item, **kwargs))
            else:
                results.append(cls(item))

        return self.__class__(results)

    def merge(self, items):
        if not isinstance(items, list):
            raise ValueError("Unable to merge uncompatible types")

        items = self.__get_items(items)

        self._items += items
        return self

    def pluck(self, value, key=None):
        if key:
            attributes = {}
        else:
            attributes = []

        if isinstance(self._items, dict):
            return Collection([self._items.get(value)])

        for item in self:
            if isinstance(item, dict):
                iterable = item.items()
            elif hasattr(item, "serialize"):
                iterable = item.serialize().items()
            else:
                iterable = self.all().items()

            for k, v in iterable:
                if k == value:
                    if key:
                        attributes[self._data_get(item, key)] = self._data_get(
                            item, value
                        )
                    else:
                        attributes.append(v)

        return Collection(attributes)

    def pop(self):
        last = self._items.pop()
        return last

    def prepend(self, value):
        self._items.insert(0, value)
        return self

    def pull(self, key):
        value = self.get(key)
        self.forget(key)
        return value

    def push(self, value):
        self._items.append(value)

    def put(self, key, value):
        self[key] = value
        return self

    def random(self, count=None):
        """Returns a random item of the collection."""
        collection_count = self.count()
        if collection_count == 0:
            return None
        elif count and count > collection_count:
            raise ValueError("count argument must be inferior to collection length.")
        elif count:
            self._items = random.sample(self._items, k=count)
            return self
        else:
            return random.choice(self._items)

    def reduce(self, callback, initial=0):
        return reduce(callback, self, initial)

    def reject(self, callback):
        self._check_is_callable(callback)

        items = self._get_value(callback) or self._items
        self._items = items

    def reverse(self):
        self._items = self[::-1]

    def serialize(self):
        def _serialize(item):
            if self.__appends__:
                item.set_appends(self.__appends__)

            if hasattr(item, "serialize"):
                return item.serialize()
            elif hasattr(item, "to_dict"):
                return item.to_dict()
            return item

        return list(map(_serialize, self))

    def add_relation(self, result=None):
        for model in self._items:
            model.add_relations(result or {})

        return self

    def shift(self):
        return self.pull(0)

    def sort(self, key=None):
        if key:
            self._items.sort(key=lambda x: x[key], reverse=False)
            return self

        self._items = sorted(self)
        return self

    def sum(self, key=None):
        result = 0
        items = self._get_value(key) or self._items
        try:
            result = sum(items)
        except TypeError:
            pass
        return result

    def to_json(self, **kwargs):
        return json.dumps(self.serialize(), **kwargs)

    def group_by(self, key):

        from itertools import groupby

        self.sort(key)

        new_dict = {}

        for k, v in groupby(self._items, key=lambda x: x[key]):
            new_dict.update({k: list(v)})

        return Collection(new_dict)

    def transform(self, callback):
        self._check_is_callable(callback)
        self._items = self._get_value(callback)

    def unique(self, key=None):
        if not key:
            items = list(set(self._items))
            return self.__class__(items)

        keys = set()
        items = []
        if isinstance(self.all(), dict):
            return self

        for item in self:
            if isinstance(item, dict):
                comparison = item.get(key)
            elif isinstance(item, str):
                comparison = item
            else:
                comparison = getattr(item, key)
            if comparison not in keys:
                items.append(item)
                keys.add(comparison)

        return self.__class__(items)

    def where(self, key, *args):
        op = "=="
        value = args[0]

        if len(args) >= 2:
            op = args[0]
            value = args[1]

        attributes = []

        for item in self._items:
            if isinstance(item, dict):
                comparison = item.get(key)
            else:
                comparison = getattr(item, key)
            if self._make_comparison(comparison, value, op):
                attributes.append(item)

        return self.__class__(attributes)

    def zip(self, items):
        items = self.__get_items(items)
        if not isinstance(items, list):
            raise ValueError("The 'items' parameter must be a list or a Collection")

        _items = []
        for x, y in zip(self, items):
            _items.append([x, y])
        return self.__class__(_items)

    def set_appends(self, appends):
        """
        Set the attributes that should be appended to the Collection.

        :rtype: list
        """
        self.__appends__ += appends
        return self

    def _get_value(self, key):
        if not key:
            return None

        items = []
        for item in self:
            if isinstance(key, str):
                if hasattr(item, key) or (key in item):
                    items.append(getattr(item, key, item[key]))
            elif callable(key):
                result = key(item)
                if result:
                    items.append(result)
        return items

    def _data_get(self, item, key, default=None):
        try:
            if isinstance(item, (list, tuple)):
                item = item[key]
            elif isinstance(item, (dict, Dotty)):
                item = data_get(item, key, default)
            elif isinstance(item, object):
                item = getattr(item, key)
        except (IndexError, AttributeError, KeyError, TypeError):
            return self._value(default)

        return item

    def _value(self, value):
        if callable(value):
            return value()
        return value

    def _check_is_callable(self, callback, raise_exception=True):
        if not callable(callback):
            if not raise_exception:
                return False
            raise ValueError("The 'callback' should be a function")
        return True

    def _make_comparison(self, a, b, op):
        operators = {
            "<": operator.lt,
            "<=": operator.le,
            "==": operator.eq,
            "!=": operator.ne,
            ">": operator.gt,
            ">=": operator.ge,
        }
        return operators[op](a, b)

    def __iter__(self):
        for item in self._items:
            yield item

    def __eq__(self, other):
        if isinstance(other, Collection):
            return other == other.all()
        return other == self._items

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self.__class__(self._items[item])

        return self._items[item]

    def __setitem__(self, key, value):
        self._items[key] = value

    def __delitem__(self, key):
        del self._items[key]

    def __ne__(self, other):
        other = self.__get_items(other)
        return other != self._items

    def __len__(self):
        return len(self._items)

    def __le__(self, other):
        other = self.__get_items(other)
        return self._items <= other

    def __lt__(self, other):
        other = self.__get_items(other)
        return self._items < other

    def __ge__(self, other):
        other = self.__get_items(other)
        return self._items >= other

    def __gt__(self, other):
        other = self.__get_items(other)
        return self._items > other

    @classmethod
    def __get_items(cls, items):
        if isinstance(items, Collection):
            items = items.all()

        return items


def collect(iterable):
    """Transform an iterable into a collection."""
    return Collection(iterable)


def flatten(iterable):
    """Flatten all sub-iterables of an iterable structure (recursively)."""
    flat_list = []
    for item in iterable:
        if isinstance(item, list):
            for subitem in flatten(item):
                flat_list.append(subitem)
        else:
            flat_list.append(item)

    return flat_list



from ..exceptions import RouteMiddlewareNotFound


class MiddlewareCapsule:
    def __init__(self):
        self.route_middleware = {}
        self.http_middleware = []

    def add(self, middleware):
        if isinstance(middleware, dict):
            self.route_middleware.update(middleware)

        if isinstance(middleware, list):
            self.http_middleware += middleware

        return self

    def remove(self, middleware):
        if middleware in self.route_middleware:
            self.route_middleware.pop(middleware)
        elif middleware in self.http_middleware:
            self.http_middleware.pop(self.http_middleware.index(middleware))
        return self

    def get_route_middleware(self, keys=None):
        middlewares = []
        if keys is None:
            return self.route_middleware

        if keys is None:
            keys = []

        for key in keys:
            try:
                found = self.route_middleware[key]
            except KeyError:
                raise RouteMiddlewareNotFound(
                    f"Could not find the '{key}' middleware key."
                )
            if isinstance(found, list):
                middlewares += found
            else:
                middlewares += [found]
        return middlewares

    def get_http_middleware(self):
        return self.http_middleware

    def run_route_middleware(self, middlewares, request, response, callback="before"):
        for middleware in middlewares:
            if ":" in middleware:
                # get list of arguments if any
                middleware_to_run, raw_arguments = middleware.split(":")
                raw_arguments = raw_arguments.split(",")
                # try to parse arguments with @ from requests
                arguments = []
                for arg in raw_arguments:
                    if "@" in arg:
                        arg = arg.replace("@", "")
                        arg = request.input(arg)
                    arguments.append(arg)
                arguments = tuple(arguments)
            else:
                middleware_to_run = middleware
                arguments = ()
            route_middlewares = self.get_route_middleware([middleware_to_run])
            for route_middleware in route_middlewares:
                middleware_response = getattr(route_middleware(), callback)(
                    request, response, *arguments
                )
                if middleware_response != request:
                    return False
        return True



from ..Broadcast import Broadcast
from ...request import Request
from ...controllers import Controller


class BroadcastingController(Controller):
    def authorize(self, request: Request, broadcast: Broadcast):
        return broadcast.driver("pusher").authorize(
            request.input("channel_name"), request.input("socket_id")
        )



from .. import Middleware
from ...utils.str import random_string
from ...facades import Request, Session, Response
from ...validation import MessageBag


class SessionMiddleware(Middleware):
    def before(self, request, response):
        if not request.cookie("SESSID"):
            session_code = random_string(10)
            response.cookie("SESSID", session_code)
            request.cookie("SESSID", session_code)
        Session.start()
        request.app.make("response").with_input = self.with_input
        request.app.make("response").with_errors = self.with_errors
        request.app.make("response").with_success = self.with_success
        request.app.make("request").session = Session

        # TODO: Remove in Masonite 5
        request.app.make("view").share({"bag": MessageBag(Session.get("errors")).helper})
        return request

    def after(self, request, _):
        return request

    def with_input(self):
        for key, value in Request.all().items():
            Session.flash(key, value)
        return Response

    def with_errors(self, errors):
        if isinstance(errors, list):
            Session.flash("errors", {"errors": errors})
        else:
            Session.flash("errors", errors)

        return Response

    def with_success(self, success):
        if isinstance(success, list):
            Session.flash("success", {"success": success})
        else:
            Session.flash("success", success)

        return Response



from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ..foundation import Application


class Storage:
    """File storage manager for Masonite handling managing files with different drivers."""

    def __init__(self, application: "Application", store_config: dict = None):
        self.application = application
        self.drivers = {}
        self.store_config = store_config or {}
        self.options = {}

    def add_driver(self, name: str, driver: str):
        self.drivers.update({name: driver})

    def set_configuration(self, config: dict) -> "Storage":
        self.store_config = config
        return self

    def get_driver(self, name: str = None) -> Any:
        if name is None:
            return self.drivers[self.store_config.get("default")]
        return self.drivers[name]

    def get_config_options(self, name: str = None) -> dict:
        if name is None or name == "default":
            return self.store_config.get(self.store_config.get("default"))

        return self.store_config.get(name)

    def disk(self, name: str = "default") -> Any:
        """Get the file manager instance for the given disk name."""
        store_config = self.get_config_options(name)
        driver = self.get_driver(self.get_config_options(name).get("driver"))
        return driver.set_options(store_config)



"""Broadcast notification driver."""

from .BaseDriver import BaseDriver


class BroadcastDriver(BaseDriver):
    def __init__(self, application):
        self.application = application
        self.options = {}

    def set_options(self, options):
        self.options = options
        return self

    def send(self, notifiable, notification):
        """Used to broadcast a notification."""
        data = self.get_data("broadcast", notifiable, notification)
        channels = notification.broadcast_on() or notifiable.route_notification_for(
            "broadcast"
        )
        event = notification.type()
        self.application.make("broadcast").channel(channels, event, data)



from .Task import Task
import subprocess


class CommandTask(Task):

    run_every_minute = True

    def __init__(self, command=""):
        self.command = command

    def handle(self):
        subprocess.call(self.command.split(" "))



"""Scaffold Auth Command."""
from shutil import copytree
import os
import sys

from ..utils.location import controllers_path, views_path, mailables_path
from ..utils.filesystem import get_module_dir
from .Command import Command


class AuthCommand(Command):
    """
    Creates a new authentication scaffold.

    auth
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        # dirs_exist_ok option is available in Python 3.8+
        # https://docs.python.org/3/library/shutil.html#shutil.copytree
        if sys.version_info.minor >= 8:
            copytree(self.get_template_path(), views_path("auth"), dirs_exist_ok=True)
            copytree(
                self.get_controllers_path(),
                controllers_path("auth"),
                dirs_exist_ok=True,
            )

            copytree(self.mailables_path(), mailables_path(), dirs_exist_ok=True)
        else:
            copytree(self.get_template_path(), views_path("auth"))
            copytree(self.get_controllers_path(), controllers_path("auth"))
            copytree(self.mailables_path(), mailables_path())

        self.info("Auth scaffolded successfully!")

    def get_template_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/templates/auth")

    def get_controllers_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/controllers/auth")

    def mailables_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/auth/mailables")



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...request import Request
    from ...response import Response

from ...utils.str import match


class CorsMiddleware:
    def before(self, request: "Request", response: "Response"):
        cors = request.app.make("cors")

        if not self.match_paths(request, cors):
            return request

        if cors.is_preflight(request):
            response = cors.build_preflight_response(request, response)
            response = cors.update_vary_header(
                response, "Access-Control-Request-Method"
            )
            return response

        return request

    def after(self, request: "Request", response: "Response"):
        cors = request.app.make("cors")

        if request.get_request_method() == "OPTIONS":
            cors.update_vary_header(response, "Access-Control-Request-Method")

        response = cors.add_actual_request_headers(request, response)
        return response

    def match_paths(self, request: "Request", cors):
        host = request.get_host()
        request_path = request.get_path().strip("/")
        paths = cors.options.get("paths")

        for path in paths:
            if path != "/":
                path = path.strip("/")

            if path == host:
                return True
            elif match(request_path, path):
                return True

        return False



"""Time related helpers"""
import pendulum


def cookie_expire_time(str_time):
    """Take a string like 1 month or 5 minutes and returns a datetime formatted with cookie format.

    Arguments:
        str_time {string} -- Could be values like 1 second or 3 minutes

    Returns:
        str -- Cookie expiration time (Thu, 21 Oct 2021 07:28:00)
    """
    instance = parse_human_time(str_time)
    return instance.format("ddd, DD MMM YYYY HH:mm:ss")


def parse_human_time(str_time):
    """Take a string like 1 month or 5 minutes and returns a pendulum instance.

    Arguments:
        str_time {string} -- Could be values like 1 second or 3 minutes

    Returns:
        pendulum -- Returns Pendulum instance
    """
    if str_time == "now":
        return pendulum.now("GMT")

    if str_time != "expired":
        number = int(str_time.split(" ")[0])
        length = str_time.split(" ")[1]

        if length in ("second", "seconds"):
            return pendulum.now("GMT").add(seconds=number)
        elif length in ("minute", "minutes"):
            return pendulum.now("GMT").add(minutes=number)
        elif length in ("hour", "hours"):
            return pendulum.now("GMT").add(hours=number)
        elif length in ("day", "days"):
            return pendulum.now("GMT").add(days=number)
        elif length in ("week", "weeks"):
            return pendulum.now("GMT").add(weeks=number)
        elif length in ("month", "months"):
            return pendulum.now("GMT").add(months=number)
        elif length in ("year", "years"):
            return pendulum.now("GMT").add(years=number)

        return None
    else:
        return pendulum.now("GMT").subtract(years=20)


def migration_timestamp():
    """Return current time formatted for creating migration filenames.
    Example: 2021_01_09_043202
    """
    return pendulum.now().format("YYYY_MM_DD_HHmmss")



from ...providers import Provider
from ..Storage import Storage
from ...configuration import config
from ..drivers import LocalDriver, AmazonS3Driver


class StorageProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        storage = Storage(self.application).set_configuration(
            config("filesystem.disks")
        )
        storage.add_driver("file", LocalDriver(self.application))
        storage.add_driver("s3", AmazonS3Driver(self.application))
        self.application.bind("storage", storage)

    def boot(self):
        pass



class Cache:
    def __init__(self, application, store_config=None):
        self.application = application
        self.drivers = {}
        self.store_config = store_config or {}
        self.options = {}

    def add_driver(self, name, driver):
        self.drivers.update({name: driver})

    def set_configuration(self, config):
        self.store_config = config
        return self

    def get_driver(self, name=None):
        if name is None:
            return self.drivers[self.store_config.get("default")]
        return self.drivers[name]

    def get_config_options(self, name=None):
        if name is None or name == "default":
            return self.store_config.get(self.store_config.get("default"))

        return self.store_config.get(name)

    def store(self, name="default"):
        store_config = self.get_config_options(name)
        driver = self.get_driver(self.get_config_options(name).get("driver"))
        return driver.set_options(store_config)

    def add(self, *args, store=None, **kwargs):
        return self.store(name=store).add(*args, **kwargs)

    def get(self, *args, store=None, **kwargs):
        return self.store(name=store).get(*args, **kwargs)

    def put(self, *args, store=None, **kwargs):
        return self.store(name=store).put(*args, **kwargs)

    def has(self, *args, store=None, **kwargs):
        return self.store(name=store).has(*args, **kwargs)

    def forget(self, *args, store=None, **kwargs):
        return self.store(name=store).forget(*args, **kwargs)

    def increment(self, *args, store=None, **kwargs):
        return self.store(name=store).increment(*args, **kwargs)

    def decrement(self, *args, store=None, **kwargs):
        return self.store(name=store).decrement(*args, **kwargs)

    def flush(self, *args, store=None, **kwargs):
        return self.store(name=store).flush(*args, **kwargs)

    def remember(self, *args, store=None, **kwargs):
        return self.store(name=store).remember(*args, **kwargs)



import inspect
import os

from ...utils.filesystem import get_module_dir


def is_property(obj):
    return not inspect.ismethod(obj)


def is_local(obj_name, obj):
    return (
        not obj_name.startswith("__")
        and not obj_name.endswith("__")
        and not type(obj).__name__ == "builtin_function_or_method"
    )


def serialize_property(obj):
    if isinstance(obj, list):
        local_list = []
        for subobj in obj:
            local_list.append(serialize_property(subobj))
        return local_list
    elif isinstance(obj, dict):
        local_dict = {}
        for key, val in obj.items():
            local_dict.update({key: serialize_property(val)})
        return local_dict
    elif hasattr(obj, "serialize"):
        return obj.serialize()
    else:
        return str(obj)


class DumpExceptionHandler:
    def __init__(self, application):
        self.application = application

        self.assets_path = os.path.join(
            get_module_dir(__file__), "../../templates/assets"
        )
        self.styles = []
        self.scripts = []

    def add_style(self, file):
        with open(os.path.join(self.assets_path, file), "r") as f:
            self.styles.append(f.read())

    def add_script(self, file):
        with open(os.path.join(self.assets_path, file), "r") as f:
            self.scripts.append(f.read())

    def get_scripts(self):
        scripts_str = ""
        for script in self.scripts:
            scripts_str += f"<script>{script}</script>\n"
        return scripts_str

    def get_styles(self):
        styles_str = ""
        for style in self.styles:
            styles_str += f"<style>{style}</style>\n"
        return styles_str

    def handle(self, exception):
        dumps = []
        # for dump in self.application.make("dumper").get_dumps():
        # for obj_name, obj in dump.objects.items():
        # all_members = inspect.getmembers(obj, predicate=inspect.ismethod)
        # all_properties = inspect.getmembers(obj, predicate=is_property)
        # members = {
        #     name: str(member)
        #     for name, member in all_members
        #     if is_local(name, member)
        # }
        # properties = {
        #     name: serialize_property(prop)
        #     for name, prop in all_properties
        #     if is_local(name, prop)
        # }
        # dumps.append(
        #     {
        #         "name": obj_name,
        #         "obj": str(obj),
        #         "members": members,
        #         "properties": properties,
        #     }
        # )
        dumps = self.application.make("dumper").get_serialized_dumps()
        self.add_style("tailwind.css")
        self.add_style("github-dark.min.css")
        self.add_script("highlight.min.js")

        return self.application.make("response").view(
            self.application.make("view").render(
                "/masonite/templates/dump",
                {
                    "styles": self.get_styles(),
                    "scripts": self.get_scripts(),
                    "dumps": dumps,
                },
            )
        )



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from masonite.foundation import Application


class BaseDriver:
    """Base Session driver"""

    def __init__(self, application: "Application"):
        self.application = application
        self.options = {}

    def set_options(self, options: dict):
        self.options = options
        return self



from inspect import isclass

from ..facades import Response as ResponseFacade
from .Provider import Provider
from ..routes import Route
from ..routes.commands import RouteListCommand
from ..pipeline import Pipeline
from ..exceptions import RouteNotFoundException, MethodNotAllowedException


class RouteProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        Route.set_controller_locations(self.application.make("controllers.location"))
        self.application.make("commands").add(RouteListCommand(self.application))

    def boot(self):
        from ..response import Response

        router = self.application.make("router")
        request = self.application.make("request")
        response = self.application.make("response")

        route = router.find(
            request.get_path(), request.get_request_method(), request.get_subdomain()
        )
        if not route:
            raise RouteNotFoundException(
                f"{request.get_request_method()} {request.get_path()} : 404 Not Found"
            )

        request.route = route

        # Run before middleware
        before_middleware = Pipeline(request, response).through(
            self.application.make("middleware").get_http_middleware(),
            handler="before",
        )

        exception = None
        if before_middleware:
            request.load_params(route.extract_parameters(request.get_path()))
            route_middleware = self.application.make("middleware").run_route_middleware(
                route.get_middlewares(), request, response, callback="before"
            )

            if route_middleware:
                try:
                    data = route.get_response(self.application)
                    if isinstance(data, Response) or (
                        isclass(data) and issubclass(data, ResponseFacade)
                    ):
                        pass
                    else:
                        response.view(data)
                except Exception as e:
                    exception = e

                self.application.make("middleware").run_route_middleware(
                    route.get_middlewares(), request, response, callback="after"
                )

        Pipeline(request, response).through(
            self.application.make("middleware").get_http_middleware(),
            handler="after",
        )

        if exception:
            raise exception



"""Vonage notification driver."""
from ....exceptions import NotificationException
from ..BaseDriver import BaseDriver


class VonageDriver(BaseDriver):
    def __init__(self, application):
        self.app = application
        self.options = {}

    def set_options(self, options):
        self.options = options
        return self

    def build(self, notifiable, notification):
        """Build SMS payload sent to Vonage API."""
        sms = self.get_data("vonage", notifiable, notification)
        if not sms._from:
            sms = sms.from_(self.options.get("sms_from"))
        if not sms._to:
            recipients = notifiable.route_notification_for("vonage")
            sms = sms.to(recipients)
        return sms

    def get_sms_client(self):
        try:
            import vonage
            from vonage.sms import Sms
        except ImportError:
            raise ModuleNotFoundError(
                "Could not find the 'vonage' library. Run 'pip install vonage' to fix this."
            )
        client = vonage.Client(
            key=self.options.get("key"), secret=self.options.get("secret")
        )
        return Sms(client)

    def send(self, notifiable, notification):
        """Used to send the SMS."""
        sms = self.build(notifiable, notification)
        client = self.get_sms_client()
        recipients = sms._to
        for recipient in recipients:
            payload = sms.to(recipient).build().get_options()
            response = client.send_message(payload)
            self._handle_errors(response)
        return response

    def _handle_errors(self, response):
        """Handle errors of Vonage API. Raises VonageAPIError if request does
        not succeed.

        An error message is structured as follows:
        {'message-count': '1', 'messages': [{'status': '2', 'error-text': 'Missing api_key'}]}
        As a success message can be structured as follows:
        {'message-count': '1', 'messages': [{'to': '3365231278', 'message-id': '140000012BD37332', 'status': '0',
        'remaining-balance': '1.87440000', 'message-price': '0.06280000', 'network': '20810'}]}

        More informations on status code errors: https://developer.nexmo.com/api-errors/sms

        """
        for message in response.get("messages", []):
            status = message["status"]
            if status != "0":
                raise NotificationException(
                    "Vonage Code [{0}]: {1}. Please refer to API documentation for more details.".format(
                        status, message["error-text"]
                    )
                )



"""New Notification Command"""
import inflection
import os

from ...utils.filesystem import get_module_dir, make_directory, render_stub_file
from ...utils.location import base_path
from ...utils.str import as_filepath
from ...commands.Command import Command


class MakeNotificationCommand(Command):
    """
    Creates a new notification class.

    notification
        {name : Name of the notification}
        {--f|force=? : Force overriding file if already exists}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        name = inflection.camelize(self.argument("name"))

        content = render_stub_file(self.get_stub_notification_path(), name)

        relative_filename = os.path.join(
            as_filepath(self.app.make("notifications.location")), name + ".py"
        )
        filepath = base_path(relative_filename)
        make_directory(filepath)
        if os.path.exists(filepath) and not self.option("force"):
            self.warning(
                f"{filepath} already exists! Run the command with -f (force) to override."
            )
            return -1
        with open(filepath, "w") as f:
            f.write(content)

        self.info(f"Notification Created ({relative_filename})")

    def get_stub_notification_path(self):
        return os.path.join(
            get_module_dir(__file__), "../../stubs/notification/Notification.py"
        )



"""New Queue Table Command."""
import os

from ..utils.filesystem import make_directory, get_module_dir
from ..utils.time import migration_timestamp
from ..utils.location import base_path
from .Command import Command


class QueueTableCommand(Command):
    """
    Creates the jobs table

    queue:table
        {--d|--directory=databases/migrations : Specifies the directory to create the migration in}
    """

    def handle(self):
        with open(
            os.path.join(
                get_module_dir(__file__), "../stubs/queue/create_queue_jobs_table.py"
            )
        ) as fp:
            output = fp.read()

        relative_filename = os.path.join(
            self.option("directory"),
            f"{migration_timestamp()}_create_queue_jobs_table.py",
        )
        filepath = base_path(relative_filename)
        make_directory(filepath)

        with open(filepath, "w") as fp:
            fp.write(output)

        self.info(f"Migration file created: {relative_filename}")



import json
import io
import os
from pprint import pprint
import sys
import pytest
import unittest
import pendulum
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any
from pprint import pprint


if TYPE_CHECKING:
    from ..foundation import Application
    from ..routes import HTTPRoute
    from .HttpTestResponse import HttpTestResponse
    from pendulum import DateTime
    from masoniteorm.models import Model

from ..cookies import CookieJar
from ..routes import Route
from ..utils.http import generate_wsgi
from ..request import Request
from ..headers import HeaderBag, Header
from ..response import Response
from ..environment import LoadEnvironment
from ..facades import Config, Session
from ..providers.RouteProvider import RouteProvider
from ..providers.FrameworkProvider import FrameworkProvider
from ..exceptions import RouteNotFoundException
from .TestCommand import TestCommand


class TestCase(unittest.TestCase):
    routes_to_restore = ()

    def setUp(self):
        """Define code that should be run before each unit tests."""
        LoadEnvironment("testing")
        from wsgi import application

        self.application: "Application" = application
        self.original_class_mocks = {}

        self._acting_as = {}
        self._test_cookies = {}
        self._test_headers = {}
        self._test_session = {}
        self._console_out = None
        self._console_err = None

        if hasattr(self, "startTestRun"):
            self.startTestRun()

        if hasattr(self, "connection") and self.connection:
            self.application.make("resolver")._connection_details["default"] = self.connection

        self.withoutCsrf()

        self.withoutExceptionsHandling()
        # boot providers as they won't not be loaded if the test is not doing a request
        self.application.bind("environ", generate_wsgi())
        try:
            for provider in self.application.get_providers():
                # if no request is made we don't need RouteProvider, and we can't load it anyway
                # because we don't have created a CSRF token yet
                if not isinstance(provider, RouteProvider):
                    application.resolve(provider.boot)
        except Exception as e:
            if not self._exception_handling:
                raise e
            self.application.make("exception_handler").handle(e)

        self.routes_to_restore = set(self.application.make("router").routes)

    def tearDown(self):
        """Define code that should be run after each unit tests."""
        self.withoutCsrf()
        self.withoutExceptionsHandling()
        self._acting_as = {}
        self._test_cookies = {}
        self._test_session = {}
        self._test_headers = {}

        # restore routes
        if self.routes_to_restore:
            self.application.make("router").routes = list(self.routes_to_restore)
        if hasattr(self, "stopTestRun"):
            self.stopTestRun()

        # restore console output
        self._console_out = None
        self._console_err = None

        # logout users
        self.application.make("auth").logout()

    @pytest.fixture(autouse=True)
    def _pass_fixtures(self, capsys):
        """Add all useful pytest fixtures to unittest.
        In the future, when needed more pytest fixtures could be integrated to Masonite TestCase.
        """
        # 'capsys' fixture allow to read output/error from stdout/stderr
        self.capsys = capsys

    def _readConsoleOutput(self):
        """Read console output if it has not been read yet."""
        if self._console_out is None and self._console_err is None:
            output = self.capsys.readouterr()
            self._console_out = output.out
            self._console_err = output.err

    def assertConsoleEmpty(self) -> "TestCase":
        """Assert that nothing (output or error) has been printed to the console."""
        self._readConsoleOutput()
        self.assertEqual("", self._console_out)
        self.assertEqual("", self._console_err)
        return self

    def assertConsoleNotEmpty(self) -> "TestCase":
        """Assert that something (output or error) has been printed to the console."""
        self._readConsoleOutput()
        assert self._console_out != "" or self._console_err != ""
        return self

    def assertConsoleExactOutput(self, output: str) -> "TestCase":
        """Assert that console standard output is equal to given output."""
        self._readConsoleOutput()
        self.assertEqual(output, self._console_out)
        return self

    def assertConsoleOutputContains(self, output: str) -> "TestCase":
        """Assert that console standard output contains given output."""
        self._readConsoleOutput()
        self.assertIn(output, self._console_out)
        return self

    def assertConsoleOutputMissing(self, output: str) -> "TestCase":
        """Assert that console standard output does not contain the given output."""
        self._readConsoleOutput()
        self.assertNotIn(output, self._console_out)
        return self

    def assertConsoleHasErrors(self) -> "TestCase":
        """Assert that something has been output to console standard error."""
        self._readConsoleOutput()
        self.assertNotEqual(self._console_err, "")
        return self

    def assertConsoleExactError(self, error: str) -> "TestCase":
        """Assert that console standard error is equal to given error."""
        self._readConsoleOutput()
        self.assertEqual(error, self._console_err)
        return self

    def assertConsoleErrorContains(self, error: str) -> "TestCase":
        """Assert that console standard error contains given error."""
        self._readConsoleOutput()
        self.assertIn(error, self._console_err)
        return self

    def withExceptionsHandling(self) -> None:
        """Enable for the duration of a test the handling of exceptions through Masonite exception
        handler."""
        self._exception_handling = True

    def withoutExceptionsHandling(self) -> None:
        """Disable handling of exceptions with Masonite exception handler. Exceptions will be
        raised directly in the tests. This is the default behaviour."""
        self._exception_handling = False

    def setRoutes(self, *routes: "HTTPRoute") -> None:
        """Set all routes of router during lifetime of a test."""
        self.application.make("router").set(Route.group(*routes, middleware=["web"]))
        return self

    def addRoutes(self, *routes: "HTTPRoute") -> None:
        """Add routes to router during lifetime of a test."""
        self.application.make("router").add(Route.group(*routes, middleware=["web"]))
        return self

    def withCsrf(self) -> None:
        """Enable CSRF verification during lifetime of a test"""
        self._csrf = True
        return self

    def withoutCsrf(self) -> None:
        """Disable CSRF verification during lifetime of a test. This is the default behaviour."""
        self._csrf = False
        return self

    def get(self, route: str, data=None):
        """Make a GET request route"""
        return self.fetch(route, data, method="GET")

    def post(self, route: str, data=None):
        return self.fetch(route, data, method="POST")

    def put(self, route: str, data=None):
        return self.fetch(route, data, method="PUT")

    def patch(self, route: str, data=None):
        return self.fetch(route, data, method="PATCH")

    def head(self, route, data=None):
        return self.fetch(route, data, method="HEAD")

    def options(self, route, data=None):
        return self.fetch(route, data, method="OPTIONS")

    def make_request(
        self,
        wsgi_data: dict = {},
        post_data: dict = {},
        path: str = "/",
        query_string: str = "application=Masonite",
        method: str = "GET",
    ) -> "Request":

        wsgi_environ = {
            **wsgi_data,
            "wsgi.input": io.BytesIO(bytes(json.dumps(post_data), "utf-8")),
            "CONTENT_LENGTH": len(str(json.dumps(post_data))),
        }
        request = Request(generate_wsgi(wsgi_environ, path, query_string, method))
        request.app = self.application

        self.application.bind("request", request)
        return request

    def make_response(self, data: dict = {}) -> "Response":
        response = Response(generate_wsgi(data))
        response.app = self.application

        self.application.bind("response", response)
        return response

    def fetch(
        self, path: str, data: dict = None, method: str = None
    ) -> "HttpTestResponse":
        """Run an HTTP request and get a test response on which assertions can be run."""
        # prepare WSGI request environment
        if data is None:
            data = {}

        if not self._csrf:
            token = self.application.make("sign").sign("cookie")
            data.update({"__token": "cookie"})
            http_cookie = f"SESSID={token}; csrf_token={token}"

        environ = generate_wsgi(
            {
                "CONTENT_LENGTH": len(str(json.dumps(data))),
                "wsgi.input": io.BytesIO(bytes(json.dumps(data), "utf-8")),
                "HTTP_COOKIE": http_cookie,
            },
            path=path,
            method=method,
        )
        self.application.bind("environ", environ)

        # boot all WSGI providers
        try:
            for provider in self.application.get_providers():
                if isinstance(provider, FrameworkProvider):
                    self.application.resolve(provider.boot)
                    # now we have a request, response object
                    request = self.application.make("request")

                    # add eventual cookies added inside the test (not encrypted to be able to assert value ?)
                    for name, value in self._test_cookies.items():
                        request.cookie(name, value)

                    # add eventual headers added inside the test
                    for name, value in self._test_headers.items():
                        request.header(name, value)

                    # @dev: this is impossible for now because session is started in the session middleware only...
                    # add eventual session data added inside the test
                    # for name, value in self._test_session.items():
                    #     Session.set(name, value)

                    # log user if required
                    if self._acting_as:
                        user = self._acting_as.get("user")
                        guard = self._acting_as.get("guard")
                        self.application.make("auth").guard(guard).attempt_by_id(
                            user.get_primary_key_value()
                        )
                else:
                    self.application.resolve(provider.boot)

        except Exception as e:
            if not self._exception_handling:
                raise e
            self.application.make("exception_handler").handle(e)

        response = self.application.make("response")

        self.mock_start_response(
            response.get_status_code(),
            response.get_headers() + response.cookie_jar.render_response(),
        )

        route = self.application.make("router").find(path, method)
        if route:
            return self.application.make("tests.response").build(
                self, self.application, request, response, route
            )

        exception = RouteNotFoundException(f"No route found for url {path}")
        if self._exception_handling:
            response = self.application.make("exception_handler").handle(exception)
            return self.application.make("tests.response").build(
                self, self.application, request, response, route
            )
        else:
            raise exception

    def mock_start_response(self, *args, **kwargs):
        pass

    @contextmanager
    def captureOutput(self):
        """Capture standard output (stdout/stderr) of a Masonite command."""
        new_out, new_err = io.StringIO(), io.StringIO()
        # save normal system output
        old_out, old_err = sys.stdout, sys.stderr
        try:
            sys.stdout, sys.stderr = new_out, new_err
            yield sys.stdout
        finally:
            # restore system output
            sys.stdout, sys.stderr = old_out, old_err

    @contextmanager
    def debugMode(self, enabled: bool = True):
        """Set application debug mode."""
        old_debug_mode = Config.get("application.debug")
        Config.set("application.debug", enabled)
        yield
        Config.set("application.debug", old_debug_mode)

    @contextmanager
    def env(self, environment: str):
        """Set application environment."""
        old_env = os.getenv("APP_ENV", "")
        os.environ["APP_ENV"] = environment
        yield
        os.environ["APP_ENV"] = old_env

    def craft(self, command: str, arguments_str: str = "") -> "TestCommand":
        """Run a given command in tests and obtain a TestCommand instance to assert command
        outputs.
        Example:
            self.craft("controller", "Welcome").assertSuccess()
        """
        return TestCommand(self.application).run(command, arguments_str)

    def fake(self, binding: "str") -> Any:
        """Mock a service in the container with its mocked implementation or with a given custom
        one."""

        # save original first
        self.original_class_mocks.update(
            {binding: self.application.make(binding, self.application)}
        )
        # mock by overriding with mocked version
        mock = self.application.make(f"mock.{binding}", self.application)
        if hasattr(mock, "reset"):
            mock.reset()
        self.application.bind(binding, mock)
        return mock

    def withCookies(self, cookies_dict: dict) -> "TestCase":
        """Add cookies to the request during the lifetime of the test."""
        self._test_cookies = cookies_dict
        return self

    def withHeaders(self, headers_dict: dict) -> "TestCase":
        """Add headers to the request during the lifetime of the test."""
        self._test_headers = headers_dict
        return self

    # def withSession(self, session_dict: dict) -> "TestCase":
    #     """Add session data to the request during the lifetime of the test."""
    #     self._test_session = session_dict
    #     return self

    def actingAs(self, user, guard="web") -> "TestCase":
        """Connect as the given user during the lifetime of the test. You can select the auth
        guard to use to authenticate."""
        self.application.bind("test_user", user)
        self._acting_as = {"user": user, "guard": guard}
        return self

    def actingAsGuest(self) -> "TestCase":
        """Connect as an unauthenticated user during the lifetime of the test."""
        self.application.make("auth").logout()
        self._acting_as = {}
        return self

    def restore(self, binding: str) -> None:
        """Restore the service previously mocked to the original one."""
        original = self.original_class_mocks.get(binding)
        self.application.bind(binding, original)

    def fakeTime(self, pendulum_datetime: "DateTime" = None) -> "DateTime":
        """Set a given pendulum instance to be returned when a "now" (or "today", "tomorrow",
        "yesterday") instance is created. It's really useful during tests to check
        timestamps logic."""
        if pendulum_datetime is None:
            pendulum_datetime = pendulum.now()
        pendulum.set_test_now(pendulum_datetime)
        return pendulum_datetime

    def fakeTimeTomorrow(self) -> None:
        """Set the mocked time as tomorrow."""
        self.fakeTime(pendulum.tomorrow())

    def fakeTimeYesterday(self) -> None:
        """Set the mocked time as yesterday."""
        self.fakeTime(pendulum.yesterday())

    def fakeTimeInFuture(self, offset: int, unit: str = "days") -> None:
        """Set the mocked time as an offset of days in the future. Unit can be specified
        among pendulum units: seconds, minutes, hours, days, weeks, months, years."""
        self.restoreTime()
        datetime = pendulum.now().add(**{unit: offset})
        self.fakeTime(datetime)

    def fakeTimeInPast(self, offset: int, unit: str = "days") -> None:
        """Set the mocked time as an offset of days in the past. Unit can be specified
        among pendulum units: seconds, minutes, hours, days, weeks, months, years."""
        self.restoreTime()
        datetime = pendulum.now().subtract(**{unit: offset})
        self.fakeTime(datetime)

    def restoreTime(self) -> None:
        """Restore time to correct one, so that pendulum new "now" instance are corrects.
        This method will be typically called in tearDown() method of a test class."""
        # this will clear the mock
        pendulum.set_test_now()

    def assertDatabaseCount(self, table: str, count: int) -> None:
        self.assertEqual(self.application.make("builder").table(table).count(), count)

    def assertDatabaseHas(self, table: str, query_dict: dict) -> None:
        self.assertGreaterEqual(
            self.application.make("builder").table(table).where(query_dict).count(), 1
        )

    def assertDatabaseMissing(self, table: str, query_dict: dict) -> None:
        self.assertEqual(
            self.application.make("builder").table(table).where(query_dict).count(), 0
        )

    def assertDeleted(self, instance: "Model") -> None:
        self.assertFalse(
            self.application.make("builder")
            .table(instance.get_table_name())
            .where(instance.get_primary_key(), instance.get_primary_key_value())
            .get()
        )

    def assertSoftDeleted(self, instance: "Model") -> None:
        deleted_at_column = instance.get_deleted_at_column()
        self.assertTrue(
            self.application.make("builder")
            .table(instance.get_table_name())
            .where(instance.get_primary_key(), instance.get_primary_key_value())
            .where_not_null(deleted_at_column)
            .get()
        )

    def dump(self, output: str, title: str = ""):
        """Print output to console during tests. A title can be provided to be displayed at dump
        start."""
        with self.capsys.disabled():
            print("\n")
            if title:
                print(f"\033[93m> {title}:\033[0m\n")
            pprint(output, width=110)

    def stop(self, msg: str = ""):
        """Stop current test, a message can be given and will be displayed in the
        console.

        2 is the pytest exit code for user interruption.
        https://docs.pytest.org/en/7.1.x/reference/exit-codes.html
        """
        return pytest.exit(msg, 2)



"""Bootstrap Preset"""
import shutil

from ..utils.location import resources_path
from ..utils.filesystem import make_full_directory
from .Preset import Preset


class Bootstrap(Preset):
    """
    Configure the front-end scaffolding for the application to use Bootstrap
    """

    key = "bootstrap"
    packages = {
        "bootstrap": "^5.1.3",
        "resolve-url-loader": "^4.0.0",
        "sass": "^1.45.1",
        "sass-loader": "^12.4.0",
    }

    def install(self):
        """Install the preset"""
        self.update_packages(dev=True)
        self.update_webpack_mix()
        self.update_css()
        self.remove_node_modules()

    def update_css(self):
        """Create/Override an app.scss file configured for the preset."""
        make_full_directory(resources_path("css"))
        shutil.copyfile(
            self.get_template_path("_variables.scss"),
            resources_path("css/_variables.scss"),
        )
        shutil.copyfile(
            self.get_template_path("app.scss"), resources_path("css/app.scss")
        )



from .Provider import Provider
from ..mail import Mail
from ..mail.drivers import SMTPDriver, TerminalDriver, MailgunDriver
from ..utils.structures import load
from ..mail import MockMail
from ..facades import Config


class MailProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        mail = Mail(self.application).set_configuration(Config.get("mail.drivers"))
        mail.add_driver("smtp", SMTPDriver(self.application))
        mail.add_driver("mailgun", MailgunDriver(self.application))
        mail.add_driver("terminal", TerminalDriver(self.application))
        self.application.bind("mail", mail)
        mocked_mail = MockMail(self.application).set_configuration(
            Config.get("mail.drivers")
        )
        self.application.bind("mock.mail", mocked_mail)

    def boot(self):
        pass



from .Command import Command


class PublishPackageCommand(Command):
    """
    Publish package files to your project

    package:publish
        {name : Name of the package}
        {--r|--resources=? : Resources to publish in you project (config, views, migrations...)}
        {--d|--dry=? : Just show a preview of what will be published into your project}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        from ..packages import PackageProvider

        name = self.argument("name")
        selected_provider = None
        for provider in self.app.get_providers():
            if isinstance(provider, PackageProvider) and provider.package.name == name:
                selected_provider = provider
        if not selected_provider:
            self.error(f"No package has been registered under the name {name}.")
            return

        if self.option("resources"):
            resources = self.option("resources").split(",")
        else:
            resources = None
        dry = self.option("dry")
        published_resources = selected_provider.publish(resources, dry)
        if dry:
            self.info("The following files would be published:")
        else:
            self.info("The following files have been published:")
        for resource, files in published_resources.items():
            self.info("\n")
            self.info(f"{resource.capitalize()}:")
            for f in files:
                self.info(f" - {f}")



from .Provider import Provider
from ..sessions import Session
from ..drivers.session import CookieDriver, RedisDriver
from ..configuration import config
from ..sessions import old


class SessionProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        session = Session(self.application).set_configuration(config("session.drivers"))
        session.add_driver("cookie", CookieDriver(self.application))
        session.add_driver("redis", RedisDriver(self.application))
        self.application.bind("session", session)
        self.application.make("view").share({"old": old})

    def boot(self):
        pass



from masoniteorm.commands import (
    MigrateCommand,
    MigrateRollbackCommand,
    MigrateRefreshCommand,
    MigrateResetCommand,
    MakeModelCommand,
    MakeObserverCommand,
    MigrateStatusCommand,
    MakeMigrationCommand,
    MakeSeedCommand,
    SeedRunCommand,
    MakeModelDocstringCommand,
)

from ..commands import DbShellCommand
from ...providers import Provider
from ...utils.location import migrations_path, models_path


class ORMProvider(Provider):
    """Offical Masonite ORMProvider which configure all commands to use locations defined
    in container."""

    def __init__(self, application):
        self.application = application

    def register(self):
        models_dir = models_path(absolute=False)
        seeds_dir = self.application.make("seeds.location")
        migrations_dir = migrations_path(absolute=False)
        observers_dir = self.application.make("observers.location")
        self.application.make("commands").add(
            MakeMigrationCommand(directory=migrations_dir),
            MakeSeedCommand(directory=seeds_dir),
            MakeObserverCommand(directory=observers_dir),
            MigrateCommand(directory=migrations_dir),
            MigrateResetCommand(directory=migrations_dir),
            MakeModelCommand(directory=models_dir, migrations_directory=migrations_dir),
            MakeModelDocstringCommand(),
            MigrateStatusCommand(directory=migrations_dir),
            MigrateRefreshCommand(directory=migrations_dir, seed_directory=seeds_dir),
            MigrateRollbackCommand(directory=migrations_dir),
            SeedRunCommand(directory=seeds_dir),
            DbShellCommand(),
        )

    def boot(self):
        pass



import pendulum


class Task:

    run_every = False
    run_at = False
    run_every_hour = False
    run_every_minute = False
    twice_daily = False
    run_weekly = False

    _date = None

    name = ""

    def __init__(self):
        """
        Should only be on the child class. Also needs to be resolved by the container.
        """

        pass

    def every(self, time):
        self.run_every = time
        return self

    def every_minute(self):
        self.run_every = "1 minute"
        return self

    def every_15_minutes(self):
        self.run_every = "15 minutes"
        return self

    def every_30_minutes(self):
        self.run_every = "30 minutes"
        return self

    def every_45_minutes(self):
        self.run_every = "45 minutes"
        return self

    def hourly(self):
        self.run_every = "1 hour"
        return self

    def daily(self):
        self.run_every = "1 day"
        return self

    def weekly(self):
        self.run_every = "1 week"
        return self

    def monthly(self):
        self.run_every = "1 month"
        return self

    def at(self, run_time):
        self.run_at = run_time
        return self

    def at_twice(self, run_time):
        self.twice_daily = run_time
        return self

    def daily_at(self, run_time):
        return self.daily().at(run_time)

    def handle(self):
        """Fires the task"""

        pass

    def should_run(self, date=None):
        """If the task should run"""

        # set the date
        self._set_date()

        return self._verify_run()

    def _set_date(self):
        if not self._date:
            self._date = pendulum.now()
            if hasattr(self, "timezone"):
                self._date.in_timezone(self.timezone)

    def _verify_run(self):
        if self.run_every:
            length, frequency = self.run_every.split(" ")

            if frequency in ("minute", "minutes"):
                time = int(length)
                if self._date.minute == 0 or self._date.minute % time == 0 or time == 1:
                    return True

            elif frequency in ("hour", "hours"):
                time = int(length)
                if self._date.hour % time == 0 and self._date.minute == 0:
                    return True

            elif frequency in ("day", "days"):
                time = int(length)
                if self._date.day_of_year % time == 0 and (
                    self._date.hour == 0
                    and self._date.minute == 0
                    or self._verify_run_at()
                ):
                    return True
            elif frequency in ("month", "months"):
                time = int(length)
                if (
                    self._date.month % time == 0
                    and self._date.day == 1
                    and (
                        self._date.hour == 0
                        and self._date.minute == 0
                        or (self._date.day == 0 and self._verify_run_at())
                    )
                ):
                    return True

        elif self.run_at:
            return self._verify_run_at()

        if self.run_every_minute:
            return True
        elif self.run_every_hour:
            if self._date.hour / 1 == 1:
                return True
        elif self.twice_daily:
            if self._date.hour in self.twice_daily:
                return True

        return False

    def _verify_run_at(self):
        if self._date.minute < 10:
            minute = f"0{self._date.minute}"
        else:
            minute = self._date.minute

        if f"{self._date.hour}:{minute}" == self.run_at:
            return True

        return False



"""Remove Preset"""
import os

from ..utils.location import base_path, resources_path
from .Preset import Preset


class Remove(Preset):
    """Removes any defined Preset"""

    key = "base"
    removed_packages = [
        "bootstrap",
        "resolve-url-loader",
        "sass",
        "sass-loader",
        "vue",
        "vue-loader",
        "react",
        "react-dom",
        "@babel/preset-react",
        "tailwindcss",
        "autoprefixer",
    ]

    def install(self):
        """Install the preset"""
        self.update_packages(dev=True)
        self.update_css()
        self.update_js()
        self.update_webpack_mix()
        self.remove_node_modules()
        self.remove_all_presets_file()

    def remove_all_presets_file(self):
        presets_resources_files = [
            "css/_variables.scss",
            "css/app.scss",
            "js/components/HelloWorld.vue",
            "js/components/Example.js",
            "js/App.vue",
        ]
        for f in presets_resources_files:
            filepath = resources_path(f)
            if os.path.exists(filepath):
                os.remove(filepath)

        tailwind_config = base_path("tailwind.config.js")
        if os.path.exists(tailwind_config):
            os.remove(tailwind_config)



from ...providers import Provider

from ..Configuration import Configuration


class ConfigurationProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        config = Configuration(self.application)
        config.load()
        self.application.bind("config", config)

    def boot(self):
        pass



from masonite.controllers import Controller
from masonite.views import View
from masonite.request import Request
from masonite.response import Response
from masonite.authentication import Auth


class LoginController(Controller):
    def show(self, view: View):
        return view.render("auth.login")

    def store(self, request: Request, auth: Auth, response: Response):
        login = auth.attempt(request.input("username"), request.input("password"))

        if login:
            return response.redirect(name="auth.home")

        # Go back to login page
        return response.redirect(name="login").with_errors(
            ["The email or password is incorrect"]
        )

    def logout(self, auth: Auth, response: Response):
        auth.logout()
        return response.redirect(name="login")



from .middleware_capsule import MiddlewareCapsule
from .middleware import Middleware
from .route.VerifyCsrfToken import VerifyCsrfToken
from .route.SessionMiddleware import SessionMiddleware
from .route.ShareErrorsInSessionMiddleware import ShareErrorsInSessionMiddleware
from .route.EncryptCookies import EncryptCookies
from .route.LoadUserMiddleware import LoadUserMiddleware
from .route.MaintenanceModeMiddleware import MaintenanceModeMiddleware
from .route.GuardMiddleware import GuardMiddleware
from .route.ClearDumpsBetweenRequestsMiddleware import (
    ClearDumpsBetweenRequestsMiddleware,
)
from .route.ThrottleRequestsMiddleware import ThrottleRequestsMiddleware
from .route.IpMiddleware import IpMiddleware
from .route.CorsMiddleware import CorsMiddleware



from masonite.mail import Mailable


class __class__(Mailable):
    def build(self):
        return (
            self.to("user@gmail.com")
            .subject("Masonite 4")
            .from_("admin@gmail.com")
            .text("Hello from Masonite!")
            .html("<h1>Hello from Masonite!</h1>")
        )



import pickle
import pendulum
from ...utils.console import HasColoredOutput
from ...utils.time import parse_human_time
import time


class DatabaseDriver(HasColoredOutput):
    def __init__(self, application):
        self.application = application

    def set_options(self, options):
        self.options = options
        return self

    def push(self, *jobs, args=(), **kwargs):
        builder = self.get_builder()

        available_at = parse_human_time(kwargs.get("delay", "now"))

        for job in jobs:
            payload = pickle.dumps(
                {
                    "obj": job,
                    "args": args,
                    "kwargs": kwargs,
                    "callback": self.options.get("callback", "handle"),
                }
            )

            builder.create(
                {
                    "name": str(job),
                    "payload": payload,
                    "available_at": available_at.to_datetime_string(),
                    "attempts": 0,
                    "queue": self.options.get("queue", "default"),
                }
            )

    def consume(self):
        print("Listening for jobs on queue: " + self.options.get("queue", "default"))
        builder = self.get_builder()

        while True:
            time.sleep(int(self.options.get("poll", 1)))
            if self.options.get("verbosity") == "vv":
                print("Checking for available jobs .. ")
            builder = builder.new().table(self.options.get("table"))
            jobs = (
                builder.where("queue", self.options.get("queue", "default"))
                .where(
                    "available_at",
                    "<=",
                    pendulum.now(tz=self.options.get("tz", "UTC")).to_datetime_string(),
                )
                .limit(10)
                .order_by("id")
                .get()
            )

            if self.options.get("verbosity") == "vv":
                print(f"Found {len(jobs)} job(s) ")

            builder.where_in("id", [x["id"] for x in jobs]).update(
                {
                    "reserved_at": pendulum.now(
                        tz=self.options.get("tz", "UTC")
                    ).to_datetime_string()
                }
            )

            for job in jobs:
                builder.where("id", job["id"]).table(self.options.get("table")).update(
                    {
                        "ran_at": pendulum.now(
                            tz=self.options.get("tz", "UTC")
                        ).to_datetime_string(),
                    }
                )
                payload = job["payload"]
                unserialized = pickle.loads(job["payload"])
                obj = unserialized["obj"]
                args = unserialized["args"]
                callback = unserialized["callback"]

                try:
                    try:
                        getattr(obj, callback)(*args)

                    except AttributeError:
                        obj(*args)

                    self.success(
                        f"[{job['id']}][{pendulum.now(tz=self.options.get('tz', 'UTC')).to_datetime_string()}] Job Successfully Processed"
                    )
                    if self.options.get("verbosity") == "vv":
                        print(f"Successful. Deleting Job ID: {job['id']}")
                    builder.where("id", job["id"]).delete()
                except Exception as e:  # skipcq
                    self.danger(
                        f"[{job['id']}][{pendulum.now(tz=self.options.get('tz', 'UTC')).to_datetime_string()}] Job Failed"
                    )

                    job["attempts"] = int(job["attempts"])

                    if job["attempts"] + 1 < int(self.options.get("attempts", 1)):
                        builder.where("id", job["id"]).table(
                            self.options.get("table")
                        ).update(
                            {
                                "attempts": int(job["attempts"] + 1),
                            }
                        )
                    elif job["attempts"] + 1 >= int(
                        self.options.get("attempts", 1)
                    ) and not self.options.get("failed_table"):
                        # Delete the jobs
                        builder.where("id", job["id"]).table(
                            self.options.get("table")
                        ).update(
                            {
                                "attempts": job["attempts"] + 1,
                            }
                        )

                        if hasattr(obj, "failed"):
                            getattr(obj, "failed")(unserialized, str(e))

                        builder.where("id", job["id"]).table(
                            self.options.get("table")
                        ).delete()
                    elif self.options.get("failed_table"):
                        self.add_to_failed_queue_table(
                            builder, job["name"], payload, str(e)
                        )

                        if hasattr(obj, "failed"):
                            getattr(obj, "failed")(unserialized, str(e))

                        builder.where("id", job["id"]).table(
                            self.options.get("table")
                        ).delete()
                    else:
                        builder.where("id", job["id"]).table(
                            self.options.get("table")
                        ).update(
                            {
                                "attempts": job["attempts"] + 1,
                            }
                        )

    def retry(self):
        builder = self.get_builder()

        jobs = (
            builder.table(self.options.get("failed_table"))
            .where("queue", self.options.get("queue", "default"))
            .get()
        )

        if len(jobs) == 0:
            self.success("No failed jobs found.")
            return

        for job in jobs:
            builder.table("jobs").create(
                {
                    "name": str(job["name"]),
                    "payload": job["payload"],
                    "attempts": 0,
                    "available_at": pendulum.now(
                        tz=self.options.get("tz", "UTC")
                    ).to_datetime_string(),
                    "queue": job["queue"],
                }
            )
        self.success(f"Added {len(jobs)} failed job(s) back to the queue")
        builder.table(self.options.get("failed_table", "failed_jobs")).where_in(
            "id", [x["id"] for x in jobs]
        ).delete()

    def get_builder(self):
        return (
            self.application.make("builder")
            .new()
            .on(self.options.get("connection"))
            .table(self.options.get("table"))
        )

    def add_to_failed_queue_table(self, builder, name, payload, exception):
        builder.table(self.options.get("failed_table", "failed_jobs")).create(
            {
                "driver": "database",
                "queue": self.options.get("queue", "default"),
                "name": name,
                "connection": self.options.get("connection"),
                "created_at": pendulum.now(
                    tz=self.options.get("tz", "UTC")
                ).to_datetime_string(),
                "exception": exception,
                "payload": payload,
                "failed_at": pendulum.now(
                    tz=self.options.get("tz", "UTC")
                ).to_datetime_string(),
            }
        )



class Hash:
    def __init__(self, application, driver_config=None):
        self.application = application
        self.drivers = {}
        self.driver_config = driver_config or {}
        self.options = {}

    def add_driver(self, name, driver):
        self.drivers.update({name: driver})

    def set_configuration(self, config):
        self.driver_config = config
        return self

    def get_driver(self, name=None):
        if name is None:
            return self.drivers[self.driver_config.get("default")]
        return self.drivers[name]

    def get_config_options(self, driver=None):
        if driver is None:
            return self.driver_config.get(self.driver_config.get("default"), {})
        return self.driver_config.get(driver, {})

    def make(self, string, options={}, driver=None):
        """Hash a string and return as string based on configured hashing protocol."""
        return (
            self.get_driver(driver)
            .set_options(options or self.get_config_options(driver))
            .make(string)
        )

    def make_bytes(self, string, options={}, driver=None):
        """Hash a string and return as bytes based on configured hashing protocol."""
        return (
            self.get_driver(driver)
            .set_options(options or self.get_config_options(driver))
            .make_bytes(string)
        )

    def check(self, plain_string, hashed_string, options={}, driver=None):
        """Verify that a given string matches its hashed version (based on configured hashing protocol)."""
        return (
            self.get_driver(driver)
            .set_options(options or self.get_config_options(driver))
            .check(plain_string, hashed_string)
        )

    def needs_rehash(self, hashed_string, options={}, driver=None):
        """Verify that a given hash needs to be hashed again because parameters for generating
        the hash have changed."""
        return (
            self.get_driver(driver)
            .set_options(options or self.get_config_options(driver))
            .needs_rehash(hashed_string)
        )



import os

from ..utils.filesystem import get_extension


class FileStream:
    def __init__(self, stream, name=None):
        self.stream = stream
        self._name = name

    def path(self):
        return self.stream.name

    def extension(self):
        return get_extension(self._name or self.path())

    def name(self):
        return self._name or os.path.basename(self.path())



from ..routes import Route


class Broadcast:
    def __init__(self, application, store_config=None):
        self.application = application
        self.drivers = {}
        self.store_config = store_config or {}
        self.options = {}

    def add_driver(self, name, driver):
        self.drivers.update({name: driver})

    def set_configuration(self, config):
        self.store_config = config
        return self

    def get_driver(self, name=None):
        if name is None:
            return self.drivers[self.store_config.get("default")]
        return self.drivers[name]

    def driver(self, name=None):
        store_config = self.get_config_options()
        driver = self.get_driver(None)
        return driver.set_options(store_config)

    def get_config_options(self, name=None):
        if name is None or name == "default":
            return self.store_config.get(self.store_config.get("default"))

        return self.store_config.get(name)

    def channel(self, channels, event=None, value=None, driver=None):
        store_config = self.get_config_options()
        driver = self.get_driver(driver)
        if not isinstance(event, str):
            if event is None:
                event = channels
                channels = event.broadcast_on()

            value = event.broadcast_with()
            if not isinstance(channels, list):
                channels = [channels]

            for channel in channels:
                if not channel.authorized(self.application):
                    continue
                event_class = event.broadcast_as()

                driver.set_options(store_config).channel(
                    channel.name, event_class, value
                )
        else:
            if not isinstance(channels, list):
                channels = [channels]
            for channel in channels:
                driver.set_options(store_config).channel(channel, event, value)

    @classmethod
    def routes(self, auth_route="/broadcasting/authorize"):
        from .controllers import BroadcastingController

        return [
            Route.post(auth_route, BroadcastingController.authorize).name(
                "broadcasting.authorize"
            )
        ]



import os
import shutil
from subprocess import call

from .Command import Command


class InstallCommand(Command):
    """
    Installs all of Masonite's dependencies

    install
        {--no-key : If set, craft install command will not generate and store a new key}
        {--no-dev : If set, Masonite will install without dev dependencies}
        {--f|--force : Overwrite .env if exists}
    """

    def handle(self):

        if not os.path.isfile(".env") or self.option("force"):
            shutil.copy(".env-example", ".env")

        if os.path.isfile("Pipfile"):
            try:
                if not self.option("no-dev"):
                    call(["pipenv", "install", "--dev"])
                else:
                    call(["pipenv", "install"])

                if not self.option("no-key"):
                    call(["pipenv", "shell", "new", "key", "--store"])

                return
            except Exception:
                self.comment(
                    """Pipenv could not install from your Pipfile .. reverting to pip installing requirements.txt"""
                )
                call(["python", "-m", "pip", "install", "-r", "requirements.txt"])
        elif os.path.isfile("requirements.txt"):
            call(["python", "-m", "pip", "install", "-r", "requirements.txt"])
        else:
            raise OSError("Could not find a Pipfile or a requirements.txt file")
        if not self.option("no-key"):
            try:
                self.call("key", "--store")
            except Exception:
                self.error(
                    "Could not successfully install Masonite. This could happen for several reasons but likely because of how Masonite is installed on your system and you could be hitting permission issues when Masonite is fetching required modules."
                    " If you have correctly followed the installation instructions then you should try everything again but start inside an virtual environment first to avoid any permission issues. If that does not work then seek help in"
                    " the Masonite Slack channel. Links can be found on GitHub in the main Masonite repo."
                )



from masonite.authorization import Policy


class __class__(Policy):
    def create(self, user):
        return False

    def view_any(self, user):
        return False

    def view(self, user, instance):
        return False

    def update(self, user, instance):
        return False

    def delete(self, user, instance):
        return False

    def force_delete(self, user, instance):
        return False

    def restore(self, user, instance):
        return False



"""Queue Failed Command."""
import os

from ..utils.filesystem import make_directory, get_module_dir
from ..utils.time import migration_timestamp
from ..utils.location import base_path
from .Command import Command


class QueueFailedCommand(Command):
    """
    Creates a failed jobs table

    queue:failed
        {--d|--directory=databases/migrations : Specifies the directory to create the migration in}
    """

    def handle(self):
        with open(
            os.path.join(
                get_module_dir(__file__), "../stubs/queue/create_failed_jobs_table.py"
            )
        ) as fp:
            output = fp.read()

        filename = f"{migration_timestamp()}_create_failed_jobs_table.py"
        path = os.path.join(base_path(self.option("directory")), filename)
        make_directory(path)

        with open(path, "w") as fp:
            fp.write(output)

        self.info(f"Migration file created: {filename}")



class PusherDriver:
    def __init__(self, application):
        self.application = application
        self.connection = None

    def set_options(self, options):
        self.options = options
        return self

    def get_connection(self):
        try:
            import pusher
        except ImportError:
            raise ModuleNotFoundError(
                "Could not find the 'pusher' library. Run 'pip install pusher' to fix this."
            )

        if self.connection:
            return self.connection

        self.connection = pusher.Pusher(
            app_id=str(self.options.get("app_id")),
            key=self.options.get("client"),
            secret=self.options.get("secret"),
            cluster=self.options.get("cluster"),
            ssl=self.options.get("ssl"),
        )

        return self.connection

    def channel(self, channel, event, value):
        return self.get_connection().trigger(channel, event, value)

    def authorize(self, channel, socket_id):
        return self.get_connection().authenticate(channel=channel, socket_id=socket_id)



import sys


class Limit:
    _delays_map = {"minute": 1, "hour": 60, "day": 60 * 24}

    def __init__(self, max_attempts: int = 60, delay: int = 1):
        self.max_attempts = max_attempts
        self.delay = delay  # in minutes
        self.key: str = ""

    @classmethod
    def per_minute(cls, max_attempts: int, minutes: int = 1) -> "Limit":
        return cls(max_attempts, minutes)

    @classmethod
    def per_hour(cls, max_attempts: int, hours: int = 1) -> "Limit":
        return cls(max_attempts, 60 * hours)

    @classmethod
    def per_day(cls, max_attempts: int, days: int = 1) -> "Limit":
        return cls(max_attempts, 60 * 24 * days)

    @classmethod
    def from_str(cls, limit: str) -> "Limit":
        max_attempts, delay_key = limit.split("/")
        delay = cls._delays_map.get(delay_key)
        return cls(int(max_attempts), delay)

    @classmethod
    def unlimited(cls) -> "Limit":
        return cls(sys.maxsize)

    def by(self, key: str) -> "Limit":
        self.key = key
        return self

    def is_unlimited(self):
        return self.max_attempts == sys.maxsize



from ..guards import JWTGuard
from ..commands.APIInstallCommand import APIInstallCommand
from ...configuration.helpers import config
from ...providers import Provider
from ..Api import Api

from ...routes import Route
from ...utils.structures import load


class ApiProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        api = Api(self.application).set_configuration(config("api.drivers"))
        self.application.bind("api", api)
        self.application.make("router").add(
            Route.group(
                load(self.application.make("routes.api.location"), "ROUTES"),
                middleware=["api"],
                prefix="/api",
            )
        )
        self.application.make("commands").add(APIInstallCommand(self.application))
        self.application.make("auth").add_guard("jwt", JWTGuard(self.application))

    def boot(self):
        pass



from ...providers import Provider
from ..Event import Event
from ..commands import MakeListenerCommand, MakeEventCommand


class EventProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        event = Event(self.application)
        self.application.make("commands").add(
            MakeListenerCommand(self.application), MakeEventCommand(self.application)
        )
        self.application.bind("event", event)

    def boot(self):
        pass



import hashlib
import os

from ..utils.filesystem import get_extension


class UploadedFile:
    def __init__(self, filename, content):
        self.filename = filename
        self.content = content

    def extension(self):
        return get_extension(self.filename)

    @property
    def name(self):
        return self.filename

    def path_name(self):
        return f"{self.name}{self.extension()}"

    def hash_path_name(self):
        return f"{self.hash_name()}{self.extension()}"

    def stream(self):
        return self.content

    def hash_name(self):
        return hashlib.sha1(bytes(self.name(), "utf-8")).hexdigest()

    def get_content(self):
        return self.content

    @property
    def size(self):
        return len(self.content)



"""New View Command."""
import inflection
import os

from ..utils.location import views_path
from ..utils.filesystem import get_module_dir, render_stub_file, make_directory
from .Command import Command


class MakeViewCommand(Command):
    """
    Creates a new view.

    view
        {name : Name of the view}
        {--f|force=? : Force overriding file if already exists}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        name = inflection.underscore(self.argument("name"))

        content = render_stub_file(self.get_view_path(), name)

        filename = f"{name}.html"
        path = views_path(filename)
        make_directory(path)
        if os.path.exists(path) and not self.option("force"):
            self.warning(
                f"{path} already exists! Run the command with -f (force) to override."
            )
            return -1

        with open(path, "w") as f:
            f.write(content)

        self.info(f"View Created ({views_path(filename, absolute=False)})")

    def get_view_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/templates/view.html")



from ..authorization import Gate
from .Provider import Provider


class AuthorizationProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        self.application.bind("gate", Gate(self.application))

    def boot(self):
        pass



from .Provider import Provider
from ..cache import Cache
from ..cache.drivers import FileDriver, RedisDriver, MemcacheDriver
from ..configuration import config


class CacheProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        cache = Cache(self.application).set_configuration(config("cache.stores"))
        cache.add_driver("file", FileDriver(self.application))
        cache.add_driver("redis", RedisDriver(self.application))
        cache.add_driver("memcache", MemcacheDriver(self.application))
        self.application.bind("cache", cache)

    def boot(self):
        pass



"""Queue Work Command."""
from .Command import Command


class QueueWorkCommand(Command):
    """
    Creates a new queue worker to consume queue jobs

    queue:work
        {--c|--connection : Specifies the database connection if using database driver.}
        {--queue=? : The queue to listen to}
        {--d|driver=? : Specify the driver you would like to use}
        {--p|poll=? : Specify the seconds a worker should wait before fetching new jobs}
        {--attempts=? : Specify the number of times a job should be retried before it fails}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        options = {}
        options.update({"driver": self.option("driver")})
        options.update({"poll": self.option("poll") or "1"})
        options.update({"attempts": self.option("attempts") or "3"})
        options.update({"queue": self.option("queue") or "default"})

        if self.option("verbose"):
            options.update({"verbosity": "v" + self.option("verbose")})

        return self.app.make("queue").consume(options)



from exceptionite import Block

from ... import __version__
from ...helpers import optional
from ...utils.str import get_controller_name


def recursive_serializer(data):
    if isinstance(data, (int, bool, str, bytes)):
        return data
    elif isinstance(data, (list, tuple)):
        return [recursive_serializer(item) for item in data]
    elif isinstance(data, dict):
        return {key: recursive_serializer(val) for key, val in data.items()}
    elif callable(data):
        return str(data)
    elif hasattr(data, "serialize"):
        return data.serialize()
    else:
        return str(data)


class AppBlock(Block):
    id = "application"
    name = "Application"
    icon = "DesktopComputerIcon"
    has_sections = True

    def build(self):
        request = self.handler.app.make("request")
        route = request.get_route()

        data = {
            "Info": {
                "Masonite Version": __version__,
                "Environment": self.handler.app.environment(),
                "Debug": self.handler.app.is_debug(),
            }
        }

        # add app route data
        if route:
            data.update(
                {
                    "Route": {
                        "Controller": get_controller_name(route.controller),
                        "Name": route.get_name(),
                        "Middlewares": route.get_middlewares(),
                    }
                }
            )

        # add user route data
        user = request.user()
        if user:
            data.update(
                {
                    "User": {
                        "E-mail": optional(user).email,
                        "ID": optional(user).id,
                    }
                }
            )

        return data


class RequestBlock(Block):
    id = "request"
    name = "Request"
    icon = "SwitchHorizontalIcon"
    has_sections = True

    def build(self):
        request = self.handler.app.make("request")
        # serialize inputs (e.g. in case of file)
        inputs = {}
        for name, value in request.all().items():
            inputs[name] = recursive_serializer(value)
        return {
            "Parameters": {
                "Path": request.get_path(),
                "Input": inputs or None,
                "Request Method": request.get_request_method(),
            },
            "Headers": request.header_bag.to_dict(),
        }


class ConfigBlock(Block):
    id = "config"
    name = "Configuration"
    icon = "CogIcon"
    has_sections = True

    def build(self):
        data = {}
        for section, config_data in self.handler.app.make("config").all().items():
            section_name = section.title()
            data[section_name] = recursive_serializer(config_data)
        return data



"""Base Notification facade."""


class Notification:
    def via(self, notifiable):
        """Defines the notification's delivery channels."""
        raise NotImplementedError("via() method should be implemented.")

    def should_send(self):
        return True

    def ignore_errors(self):
        return False

    def broadcast_on(self):
        return "broadcast"

    @classmethod
    def type(cls):
        """Get notification type defined with class name."""
        return cls.__name__

    def dry(self):
        """Sets whether the notification should be sent or not.

        Returns:
            self
        """
        self._dry = True
        return self

    def fail_silently(self):
        """Sets whether the notification can fail silently (without raising exceptions).

        Returns:
            self
        """
        self._fail_silently = True
        return self



"""Notification Table Command."""
from cleo import Command
import os

from ...utils.filesystem import get_module_dir, make_directory
from ...utils.time import migration_timestamp
from ...utils.location import base_path


class NotificationTableCommand(Command):
    """
    Creates the notifications table needed for storing notifications in the database.

    notification:table
        {--d|--directory=databases/migrations : Specifies the directory to create the migration in}
    """

    def handle(self):
        with open(
            os.path.join(
                get_module_dir(__file__),
                "../../stubs/notification/create_notifications_table.py",
            )
        ) as fp:
            output = fp.read()

        relative_filename = os.path.join(
            self.option("directory"),
            f"{migration_timestamp()}_create_notifications_table.py",
        )
        filepath = base_path(relative_filename)
        make_directory(filepath)

        with open(filepath, "w") as fp:
            fp.write(output)

        self.info(f"Migration file created: {relative_filename}")



"""Notifiable mixin"""
from masoniteorm.relationships import has_many

from .DatabaseNotification import DatabaseNotification
from ..exceptions.exceptions import NotificationException


class Notifiable:
    """Notifiable mixin allowing to send notification to a model. It's often used with the
    User model.

    Usage:
        user.notify(WelcomeNotification())
    """

    def notify(self, notification, drivers=[], dry=False, fail_silently=False):
        """Send the given notification."""
        from wsgi import application

        return application.make("notification").send(
            self, notification, drivers, dry, fail_silently
        )

    def route_notification_for(self, driver):
        """Get the notification routing information for the given driver. If method has not been
        defined on the model: for mail driver try to use 'email' field of model."""
        # check if routing has been specified on the model
        method_name = "route_notification_for_{0}".format(driver)

        try:
            method = getattr(self, method_name)
            return method()
        except AttributeError:
            # if no method is defined on notifiable use default
            if driver == "database":
                # with database channel, notifications are saved to database
                pass
            elif driver == "mail":
                return self.email
            else:
                raise NotificationException(
                    "Notifiable model does not implement {}".format(method_name)
                )

    @has_many("id", "notifiable_id")
    def notifications(self):
        """Get all notifications sent to the model instance. Only for 'database'
        notifications."""
        return DatabaseNotification.where("notifiable_type", "users").order_by(
            "created_at", direction="DESC"
        )

    @property
    def unread_notifications(self):
        """Get the model instance unread notifications. Only for 'database'
        notifications."""
        return self.notifications.where("read_at", "==", None)

    @property
    def read_notifications(self):
        """Get the model instance read notifications. Only for 'database'
        notifications."""
        return self.notifications.where("read_at", "!=", None)



"""New Policy Command."""
import inflection
import os

from ..utils.filesystem import make_directory
from .Command import Command


class MakePolicyCommand(Command):
    """
    Creates a new policy class.

    policy
        {name : Name of the policy}
        {--m|model=? : Create a policy for a model with a set of predefined methods}
        {--f|force=? : Force overriding file if already exists}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        name = inflection.camelize(self.argument("name"))
        if not name.endswith("Policy"):
            name += "Policy"

        if self.option("model"):
            with open(self.get_model_policy_path(), "r") as f:
                content = f.read()
                content = content.replace("__class__", name)
        else:
            with open(self.get_base_policy_path(), "r") as f:
                content = f.read()
                content = content.replace("__class__", name)

        file_name = os.path.join(
            self.app.make("policies.location").replace(".", "/"), name + ".py"
        )

        make_directory(file_name)
        if os.path.exists(file_name) and not self.option("force"):
            self.warning(
                f"{file_name} already exists! Run the command with -f (force) to override."
            )
            return -1

        with open(file_name, "w") as f:
            f.write(content)

        # add class to __init__.py
        with open(os.path.join(os.path.dirname(file_name), "__init__.py"), "a") as f:
            f.write(f"from .{name} import {name}\n")

        self.info(f"Policy Created ({file_name})")

    def get_template_path(self):
        current_path = os.path.dirname(os.path.realpath(__file__))

        return os.path.join(current_path, "../stubs/templates/")

    def get_base_policy_path(self):
        current_path = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(current_path, "../stubs/policies/Policy.py")

    def get_model_policy_path(self):
        current_path = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(current_path, "../stubs/policies/ModelPolicy.py")



"""Core of the IOC Container."""

import inspect

from ..exceptions import (
    ContainerError,
    MissingContainerBindingNotFound,
    StrictContainerException,
)


class Container:
    """Core of the Service Container.

    Performs bindings and resolving of objects to and from the container.
    """

    objects = {}
    strict = False
    override = True
    resolve_parameters = {}
    remember = False
    _hooks = {
        "make": {},
        "bind": {},
        "resolve": {},
    }

    swaps = {}
    _remembered = {}

    def bind(self, name, class_obj):
        """Bind classes into the container with a key value pair.

        Arguments:
            name {string} -- Name of the key you want to bind the object to
            class_obj {object} -- The object you want to bind

        Returns:
            self
        """
        if inspect.ismodule(class_obj):
            raise StrictContainerException(
                "Cannot bind module '{}' with key '{}' into the container".format(
                    class_obj, name
                )
            )
        if self.strict and name in self.objects:
            raise StrictContainerException(
                "You cannot override a key inside a strict container"
            )

        if self.override or name not in self.objects:
            self.fire_hook("bind", name, class_obj)
            self.objects.update({name: class_obj})

        return self

    def unbind(self, name):
        """Unbind classes from the container from a key.

        Arguments:
            name {string} -- Name of the key you want to bind the object to

        Returns:
            self
        """

        if name not in self.objects:
            return False

        del self.objects[name]

        return self

    def simple(self, obj):
        """Easy way to bind classes into the container by using passing the object only.

        Automatically generates the key for the binding process.

        Arguments:
            class_obj {object} -- The object you want to bind

        Returns:
            self
        """
        self.bind(obj if inspect.isclass(obj) else obj.__class__, obj)
        return self

    def singleton(self, name, class_obj):
        """Register a shared binding in the container.

        Arguments:
            name {string} -- Name of the key you want to bind the object to
            class_obj {object} -- The object you want to bind
        """
        obj = self.resolve(class_obj)
        self.bind(name, obj)

    def make(self, name, *arguments):
        """Retrieve a class from the container by key.

        Arguments:
            name {string} -- Key in the container that you want to get.

        Raises:
            MissingContainerBindingNotFound -- Raised if the key is not in the container.

        Returns:
            object -- Returns the object that is fetched.
        """

        if name in self.objects:
            obj = self.objects[name]
            self.fire_hook("make", name, obj)
            if inspect.isclass(obj):
                obj = self.resolve(obj, *arguments)
            return obj
        elif name in self.swaps:
            return self.swaps.get(name)
        elif inspect.isclass(name):
            try:
                obj = self._find_obj(name)
                self.fire_hook("make", name, obj)
                if inspect.isclass(obj):
                    obj = self.resolve(obj, *arguments)
            except MissingContainerBindingNotFound:
                # If we don't find a bound object already in the container,
                # we can go ahead and fall back on a simple resolve method.
                # This allows resolving dependencies without explicit
                # bindings.
                obj = self.resolve(name, *arguments)
            return obj

        raise MissingContainerBindingNotFound(
            "{0} key was not found in the container".format(name)
        )

    def has(self, name):
        """Check if a key exists in the container.

        Arguments:
            name {string} -- Key you want to check for in the container

        Returns:
            bool
        """
        if isinstance(name, str):
            return name in self.objects
        else:
            try:
                self._find_obj(name)
                return True
            except MissingContainerBindingNotFound:
                return False

    def helper(self):
        """Add a helper to create builtin functions.

        Used to more simply return
        instances of this class when building helpers.

        Returns:
            self
        """
        return self

    def resolve(self, obj, *resolving_arguments):
        """Takes an object such as a function or class method and resolves it's
        parameters from objects in the container.

        Arguments:
            obj {object} -- The object you want to resolve objects from via this container.

        Returns:
            object -- The object you tried resolving but with the correct dependencies injected.
        """
        objects = []
        passing_arguments = list(resolving_arguments)
        if self.remember and obj in self._remembered:
            objects = self._remembered[obj]
            try:
                return obj(*objects)
            except TypeError as e:
                raise ContainerError(str(e))
        elif (
            self.remember
            and not passing_arguments
            and inspect.ismethod(obj)
            and "{}.{}.{}".format(
                obj.__module__, obj.__self__.__class__.__name__, obj.__name__
            )
            in self._remembered
        ):
            location = "{}.{}.{}".format(
                obj.__module__, obj.__self__.__class__.__name__, obj.__name__
            )
            objects = self._remembered[location]
            try:
                return obj(*objects)
            except TypeError as e:
                raise ContainerError(str(e))
        else:
            for _, value in self.get_parameters(obj):
                if type(value.annotation) in (str, int, dict, list, tuple) or value.annotation in (str, int, dict, list, tuple):
                    # Ignore any times a user is simply type hinting a parameter like (parameter:str or parameter:"str").
                    # In this case we don't want to resolve anything but we do want
                    # to insert any passing arguments we passed in
                    try:
                        objects.append(passing_arguments.pop(0))
                    except IndexError:
                        pass

                    continue
                if ":" in str(value):
                    try:
                        param = self._find_annotated_parameter(value)
                    except ContainerError:
                        # This allows resolving dependencies without explicit bindings.
                        # See `self.make()`.
                        param = value.annotation
                    if inspect.isclass(param):
                        if resolving_arguments:
                            param = self.resolve(param, *resolving_arguments)
                        else:
                            param = self.resolve(param)
                    objects.append(param)
                elif "self" in str(value):
                    objects.append(obj)
                elif "=" in str(value):
                    objects.append(value.default)
                elif "*" in str(value):
                    continue
                elif self.resolve_parameters:
                    objects.append(self._find_parameter(value))
                elif resolving_arguments:
                    try:
                        objects.append(passing_arguments.pop(0))
                    except IndexError:
                        raise ContainerError(
                            "Not enough dependencies passed. Resolving object needs {} dependencies.".format(
                                len(inspect.signature(obj).parameters)
                            )
                        )
                else:
                    raise ContainerError(
                        "This container is not set to resolve parameters. You can set this in the container"
                        " constructor using the 'resolve_parameters=True' keyword argument."
                    )

        if self.remember:
            if not inspect.ismethod(obj):
                self._remembered[obj] = objects
            else:
                signature = "{}.{}.{}".format(
                    obj.__module__, obj.__self__.__class__.__name__, obj.__name__
                )
                self._remembered[signature] = objects
        return obj(*objects)

    def collect(self, search):
        """Fetch a dictionary of objects using a search query.

        Arguments:
            search {string|object} -- The string or object you want to search for.

        Raises:
            AttributeError -- Thrown if there is no wildcard in the search string

        Returns:
            dict -- Returns a dictionary of collected objects and their key bindings.
        """
        providers = {}
        if isinstance(search, str):
            # Search Can Be:
            #    '*ExceptionHook'
            #    'Sentry*'
            #    'Sentry*Hook'
            for key, value in self.objects.items():
                if isinstance(key, str):
                    if search.startswith("*"):
                        if key.endswith(search.split("*")[1]):
                            providers.update({key: value})
                    elif search.endswith("*"):
                        if key.startswith(search.split("*")[0]):
                            providers.update({key: value})
                    elif "*" in search:
                        split_search = search.split("*")
                        if key.startswith(split_search[0]) and key.endswith(
                            split_search[1]
                        ):
                            providers.update({key: value})
                    else:
                        raise AttributeError(
                            "There is no '*' in your collection search"
                        )
        else:
            for provider_key, provider_class in self.objects.items():
                if (
                    inspect.isclass(provider_class)
                    and issubclass(provider_class, search)
                ) or isinstance(provider_class, search):
                    providers.update({provider_key: provider_class})

        return providers

    def _find_annotated_parameter(self, parameter):
        """Find a given annotation in the container.

        Arguments:
            parameter {object} -- The object to find in the container.

        Raises:
            ContainerError -- Thrown when the dependency is not found in the container.

        Returns:
            object -- Returns the object found in the container.
        """
        if isinstance(parameter.annotation, str):
            return
        if parameter.annotation in self.swaps:
            obj = self.swaps[parameter.annotation]
            if callable(obj):
                return self.swaps[parameter.annotation](parameter.annotation, self)
            return obj

        for _, provider_class in self.objects.items():
            if (
                parameter.annotation == provider_class
                or parameter.annotation == provider_class.__class__
            ):
                obj = provider_class
                self.fire_hook("resolve", parameter, obj)

                return obj
            elif (
                inspect.isclass(provider_class)
                and issubclass(provider_class, parameter.annotation)
                or issubclass(provider_class.__class__, parameter.annotation)
            ):
                obj = provider_class
                self.fire_hook("resolve", parameter, obj)
                return obj

        raise ContainerError(
            "The dependency with the {0} annotation could not be resolved by the container".format(
                parameter
            )
        )

    def get_parameters(self, obj):
        return inspect.signature(obj).parameters.items()

    def _find_parameter(self, keyword):
        """Find a parameter in the container.

        Arguments:
            parameter {inspect.Paramater} -- Parameter to search for.

        Raises:
            ContainerError -- Thrown when the dependency is not found in the container.

        Returns:
            object -- Returns the object found in the container
        """
        parameter = str(keyword)

        if parameter != "self" and parameter in self.objects:
            obj = self.objects[parameter]
            self.fire_hook("resolve", parameter, obj)
            return obj
        elif "=" in parameter:
            return keyword.default

        raise ContainerError(
            "The parameter dependency with the key of {0} could not be found in the container".format(
                parameter
            )
        )

    def on_bind(self, key, obj):
        """Set some listeners for when a specific key or class in binded to the container.

        Arguments:
            key {string|object} -- The key can be a string or an object to listen for
            obj {object} -- Should be a function or class method

        Returns:
            self
        """
        return self._bind_hook("bind", key, obj)

    def on_make(self, key, obj):
        """Set some listeners for when a specific key or class is made from the container.

        Arguments:
            key {string|object} -- The key can be a string or an object to listen for
            obj {object} -- Should be a function or class method

        Returns:
            self
        """
        return self._bind_hook("make", key, obj)

    def on_resolve(self, key, obj):
        """Set some listeners for when a specific key or class is resolved from the container.

        Arguments:
            key {string|object} -- The key can be a string or an object to listen for
            obj {object} -- Should be a function or class method

        Returns:
            self
        """
        return self._bind_hook("resolve", key, obj)

    def swap(self, obj, callback):
        self.swaps.update({obj: callback})
        return self

    def fire_hook(self, action, key, obj):
        """Fire a specific hook based on a key or object.

        Arguments:
            action {string} -- Should be the action to fire (bind|make|resolve)
            key {string|object} -- The key can be a string or an object to listen for
            obj {object} -- Should be a function or class method

        Returns:
            None
        """
        if (
            str(key) in self._hooks[action]
            or inspect.isclass(obj)
            and obj in self._hooks[action]
            or obj.__class__ in self._hooks[action]
        ):

            for _, hook_list in self._hooks[action].items():
                for hook_obj in hook_list:
                    hook_obj(obj, self)

    def _bind_hook(self, hook, key, obj):
        """Internal method used to abstract away the logic for binding an
        listener to the container hooks.

        Arguments:
            hook {string} -- The hook you want to listen for (bind|make|resolve)
            key {string|object} -- The key to save for the listener
            obj {object} -- Should be a function or class method

        Returns:
            self
        """
        if key in self._hooks[hook]:
            self._hooks[hook][key].append(obj)
        else:
            self._hooks[hook].update({key: [obj]})
        return self

    def _find_obj(self, obj):
        """Find an object in the container.

        Arguments:
            obj {object} -- Any object in the container

        Raises:
            MissingContainerBindingNotFound -- Raised when the object cannot be found.

        Returns:
            object -- Returns the object in the container
        """
        for _, provider_class in self.objects.items():
            if obj in (provider_class, provider_class.__class__):
                return_obj = provider_class
                self.fire_hook("resolve", obj, return_obj)
                return return_obj
            elif (
                inspect.isclass(provider_class)
                and issubclass(provider_class, obj)
                or issubclass(provider_class.__class__, obj)
            ):
                return_obj = provider_class
                self.fire_hook("resolve", obj, return_obj)
                return return_obj

        raise MissingContainerBindingNotFound(
            "The dependency with the {0} annotation could not be resolved by the container".format(
                obj
            )
        )

    def __contains__(self, obj):
        return self.has(obj)



class Queue:
    def __init__(self, application, driver_config=None):
        self.application = application
        self.drivers = {}
        self.driver_config = driver_config or {}
        self.options = {}

    def add_driver(self, name, driver):
        self.drivers.update({name: driver})

    def set_configuration(self, config):
        self.driver_config = config
        return self

    def get_driver(self, name=None):
        if name is None:
            return self.drivers[self.driver_config.get("default")]
        return self.drivers[name]

    def get_config_options(self, driver=None):
        if driver is None:
            return self.driver_config.get(self.driver_config.get("default"), {})

        return self.driver_config.get(driver, {})

    def push(self, *jobs, **options):
        driver = self.get_driver(options.get("driver"))
        config_options = self.get_config_options(options.get("driver"))
        config_options.update({"queue": options.get("queue", "default")})
        driver.set_options(config_options)
        driver.push(*jobs)

    def consume(self, options):
        driver = self.get_driver(options.get("driver"))
        config_options = self.get_config_options(options.get("driver"))
        config_options.update(options)
        options.update(self.get_config_options(options.get("driver")))
        return driver.set_options(config_options).consume()

    def retry(self, options):
        driver = self.get_driver(options.get("driver"))
        config_options = self.get_config_options(options.get("driver"))
        config_options.update(options)
        options.update(self.get_config_options(options.get("driver")))
        return driver.set_options(config_options).retry()



import os
from ...utils.filesystem import make_full_directory, modified_date
from pathlib import Path
import pendulum
import json
import glob


class FileDriver:
    def __init__(self, application):
        self.application = application

    def set_options(self, options):
        self.options = options
        if options.get("location"):
            make_full_directory(options.get("location"))
        return self

    def add(self, key, value, seconds=None):
        exists = self.get(key)
        if exists:
            return exists

        return self.put(key, str(value), seconds=seconds)

    def get(self, key, default=None, **options):
        if not self.has(key):
            return default

        modified_at = self.get_modified_at(os.path.join(self._get_directory(), key))

        with open(os.path.join(self._get_directory(), key), "r") as f:
            value = f.read()

            if modified_at.add(seconds=self.get_cache_expiration(value)).is_past():
                self.forget(key)
                return default

            value = self.get_value(value)

        return value

    def put(self, key, value, seconds=None, **options):

        time = self.get_expiration_time(seconds)

        if isinstance(value, (dict,)):
            value = json.dumps(value)

        with open(os.path.join(self._get_directory(), key), "w") as f:
            f.write(f"{time}:{value}")

        return value

    def has(self, key):
        return Path(os.path.join(self._get_directory(), key)).exists()

    def increment(self, key, amount=1):
        return self.put(key, str(int(self.get(key)) + amount))

    def decrement(self, key, amount=1):
        return self.put(key, str(int(self.get(key)) - amount))

    def remember(self, key, callable):
        value = self.get(key)

        if value:
            return value

        callable(self)

    def forget(self, key):
        try:
            os.remove(os.path.join(self._get_directory(), key))
            return True
        except FileNotFoundError:
            return False

    def flush(self):
        files = glob.glob(f"{self._get_directory()}/*")
        for f in files:
            os.remove(f)

    def _get_directory(self):
        return self.options.get("location")

    def get_modified_at(self, filename):
        return pendulum.from_timestamp(modified_date(filename))

    def get_expiration_time(self, seconds):
        if seconds is None:
            seconds = 31557600 * 10

        return seconds

    def get_value(self, value):
        value = str(value.split(":", 1)[1])
        if value.isdigit():
            return str(value)
        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError:
            return value

    def get_cache_expiration(self, value):
        return int(value.split(":", 1)[0])



"""Queue Retry Command."""
from .Command import Command


class QueueRetryCommand(Command):
    """
    Puts all failed queue jobs back onto the queue.

    queue:retry
        {--c|--connection=default : Specifies the database connection if using database driver.}
        {--queue=default : The queue to listen to}
        {--d|driver=None : Specify the driver you would like to connect to}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        driver = None if self.option("driver") == "None" else self.option("driver")

        return self.app.make("queue").retry(
            {
                "driver": driver,
                "connection": self.option("connection"),
                "queue": self.option("queue"),
            }
        )



from ..exceptions import ModelNotFoundException


class ModelNotFoundHandler:
    def __init__(self, application):
        self.application = application

    def handle(self, exception):
        masonite_exception = ModelNotFoundException(
            "No record found with the given primary key"
        )
        self.application.make("exception_handler").handle(masonite_exception)



from masonite.controllers import Controller
from masonite.views import View
from masonite.request import Request
from masonite.response import Response
from masonite.authentication import Auth
from masonite.facades import Mail
from app.mailables.ResetPassword import ResetPassword


class PasswordResetController(Controller):
    def show(self, view: View):  # Show password_reset page
        return view.render("auth.password_reset")

    def store(self, auth: Auth, request: Request, response: Response):
        email, reset_token = auth.password_reset(request.input("email"))
        if email:
            Mail.mailable(ResetPassword(token=reset_token).to(email)).send()
        return response.back().with_success(
            [
                "Email sent. Please follow the directions in your email to complete your password reset."
            ]
        )

    def change_password(self, view: View, request: Request):
        return view.render("auth.change_password", {"token": request.param("token")})

    def store_changed_password(self, auth: Auth, request: Request, response: Response):
        errors = request.validate(
            {
                "password": "required|strong|confirmed",
            }
        )

        if errors:
            return response.back().with_errors(errors)

        if auth.reset_password(request.input("password"), request.param("token")):
            return response.back().with_success(["Password Reset Successfully"])

        return response.back().with_errors(["Could not reset your password"])



"""Sms Component"""


class Sms:

    _from = ""
    _to = ""
    _text = ""
    _client_ref = ""
    _type = "text"

    def __init__(self, text=""):
        self._text = text

    def from_(self, number):
        """Set the name or number the message should be sent from. Numbers should
        be specified in E.164 format. Details can be found here:
        https://developer.nexmo.com/messaging/sms/guides/custom-sender-id"""
        self._from = number
        return self

    def text(self, text):
        self._text = text
        return self

    def to(self, to):
        self._to = to
        return self

    def set_unicode(self):
        """Set message as unicode to handle unicode characters in text."""
        self._type = "unicode"
        return self

    def client_ref(self, client_ref):
        """Set your own client reference (up to 40 characters)."""
        if len(client_ref) > 40:
            raise ValueError("client_ref should have less then 40 characters.")
        self._client_ref = client_ref
        return self

    def build(self, *args, **kwargs):
        return self

    def get_options(self):
        base_dict = {
            "to": self._to,
            "from": self._from,
            "text": self._text,
            "type": self._type,
        }
        if self._client_ref:
            base_dict.update({"client-ref": self._client_ref})
        return base_dict



class WebGuard:
    def __init__(self, application):
        self.application = application
        self.connection = None

    def set_options(self, options):
        self.options = options
        return self

    def attempt(self, username, password):
        attempt = self.options.get("model")().attempt(username, password)
        if attempt and not self.options.get("once"):
            self.application.make("response").cookie("token", attempt.remember_token)
            self.application.make("request").set_user(attempt)
            return attempt

    def get_auth_column(self, username):
        return self.options.get("model")().get_auth_column(username)

    def register(self, dictionary):
        try:
            register = self.options.get("model")().register(dictionary)
        except Exception:
            return False
        return self.attempt_by_id(register.get_id())

    def user(self):
        """Get the currently logged in user.

        Returns:
            object|bool -- Returns the current authenticated user object or False or None if there is none.
        """
        token = self.application.make("request").cookie("token")
        if token and self.options.get("model")():
            return (
                self.options.get("model")().where("remember_token", token).first()
                or False
            )

        return False

    def attempt_by_id(self, user_id):
        """Login a user by the user ID.

        Arguments:
            user_id {string|int} -- The ID of the user model record.

        Returns:
            object|False -- Returns the current authenticated user object or False or None if there is none.
        """
        attempt = self.options.get("model")().attempt_by_id(user_id)

        if attempt and not self.options.get("once"):
            self.application.make("response").cookie("token", attempt.remember_token)
            self.application.make("request").set_user(attempt)
            return attempt

        return False

    def reset_password(self, username, new_password):
        """Login a user by the user ID.

        Arguments:
            user_id {string|int} -- The ID of the user model record.

        Returns:
            object|False -- Returns the current authenticated user object or False or None if there is none.
        """
        attempt = self.options.get("model")().reset_password(username, new_password)

        if attempt:
            return attempt

        return False

    def once(self):
        """Log in the user without saving a cookie.

        Returns:
            self
        """
        self._once = True
        return self

    def logout(self):
        return self



"""Slack notification driver"""
import requests

from ...exceptions import NotificationException
from .BaseDriver import BaseDriver


class SlackDriver(BaseDriver):

    WEBHOOK_MODE = 1
    API_MODE = 2
    send_url = "https://slack.com/api/chat.postMessage"
    channel_url = "https://slack.com/api/conversations.list"

    def __init__(self, application):
        self.application = application
        self.options = {}
        self.mode = self.WEBHOOK_MODE

    def set_options(self, options):
        self.options = options
        return self

    def send(self, notifiable, notification):
        """Used to send the notification to slack."""
        slack_message = self.build(notifiable, notification)
        if slack_message._mode == self.WEBHOOK_MODE:
            self.send_via_webhook(slack_message)
        else:
            self.send_via_api(slack_message)

    def build(self, notifiable, notification):
        """Build Slack message payload sent to Slack API or through Slack webhook."""
        slack_message = self.get_data("slack", notifiable, notification)
        recipients = self.get_recipients(notifiable)
        mode = self.get_sending_mode(recipients)
        slack_message = slack_message.mode(mode)

        if mode == self.WEBHOOK_MODE:
            slack_message = slack_message.to(recipients)
        elif mode == self.API_MODE:
            slack_message = slack_message.to(recipients)
            if not slack_message._token:
                slack_message = slack_message.token(self.options.get("token"))
        return slack_message

    def get_recipients(self, notifiable):
        recipients = notifiable.route_notification_for("slack")
        if not isinstance(recipients, (list, tuple)):
            recipients = [recipients]
        return recipients

    def get_sending_mode(self, recipients):
        modes = []
        for recipient in recipients:
            if recipient.startswith("https://hooks.slack.com"):
                modes.append(self.WEBHOOK_MODE)
            else:
                modes.append(self.API_MODE)
        if len(set(modes)) > 1:
            raise NotificationException("Slack sending mode cannot be mixed.")
        return modes[0]

    def send_via_webhook(self, slack_message):
        webhook_urls = slack_message._to
        payload = slack_message.build().get_options()
        for webhook_url in webhook_urls:
            response = requests.post(
                webhook_url,
                json=payload,
                headers={"Content-Type": "application/json"},
            )
            if response.status_code != 200:
                raise NotificationException(
                    "{}. Check Slack webhooks docs.".format(response.text)
                )

    def send_via_api(self, slack_message):
        """Send Slack notification with Slack Web API as documented
        here https://api.slack.com/methods/chat.postMessage"""
        channels = slack_message._to
        for channel in channels:
            channel = self.convert_channel(channel, slack_message._token)
            # set only one recipient at a time
            slack_message.to(channel)
            payload = slack_message.build().get_options()
            response = requests.post(self.send_url, payload).json()
            if not response["ok"]:
                raise NotificationException(
                    "{}. Check Slack API docs.".format(response["error"])
                )
            else:
                return response

    def convert_channel(self, name, token):
        """Calls the Slack API to find the channel ID if not already a channel ID.

        Arguments:
            name {string} -- The channel name to find.
        """
        if "#" not in name:
            return name
        response = requests.post(self.channel_url, {"token": token}).json()
        for channel in response["channels"]:
            if channel["name"] == name.split("#")[1]:
                return channel["id"]

        raise NotificationException(
            f"The user or channel being addressed either do not exist or is invalid: {name}"
        )



from typing import Callable, List, Tuple, Iterator, TYPE_CHECKING

if TYPE_CHECKING:
    from .Application import Application
    from ..request import Request
    from ..response import Response

ResponseHandler = Callable[[str, List[Tuple]], None]


def response_handler(environ: dict, start_response: ResponseHandler) -> Iterator:
    """The WSGI Application Server.

    Arguments:
        environ {dict} -- The WSGI environ dictionary
        start_response {WSGI callable}

    Returns:
        WSGI Response
    """
    from wsgi import application

    application.bind("environ", environ)

    """Add Environ To Service Container
    Add the environ to the service container. The environ is generated by the
    the WSGI server above and used by a service provider to manipulate the
    incoming requests
    """

    # """Execute All Service Providers That Require The WSGI Server
    # Run all service provider boot methods if the wsgi attribute is true.
    # """

    try:
        for provider in application.get_providers():
            application.resolve(provider.boot)
    except Exception as e:
        application.make("exception_handler").handle(e)

    """We Are Ready For Launch
    If we have a solid response and not redirecting then we need to return
    a 200 status code along with the data. If we don't, then we'll have
    to return a 302 redirection to where ever the user would like go
    to next.
    """
    _ = application.make("request")
    response: "Response" = application.make("response")

    start_response(
        response.get_status_code(),
        response.get_headers() + response.cookie_jar.render_response(),
    )

    """Final Step
    This will take the data variable from the Service Container and return
    it to the WSGI server.
    """
    return iter([response.get_response_content()])


def testcase_handler(
    application: "Application",
    environ: dict,
    start_response: ResponseHandler,
    exception_handling: bool = True,
) -> Tuple["Request", "Response"]:
    """The WSGI Application Server."""
    from wsgi import application

    application.bind("environ", environ)

    """Add Environ To Service Container
    Add the environ to the service container. The environ is generated by the
    the WSGI server above and used by a service provider to manipulate the
    incoming requests
    """

    # """Execute All Service Providers That Require The WSGI Server
    # Run all service provider boot methods if the wsgi attribute is true.
    # """

    try:
        for provider in application.get_providers():
            application.resolve(provider.boot)
    except Exception as e:
        if not exception_handling:
            raise e
        application.make("exception_handler").handle(e)

    """We Are Ready For Launch
    If we have a solid response and not redirecting then we need to return
    a 200 status code along with the data. If we don't, then we'll have
    to return a 302 redirection to where ever the user would like go
    to next.
    """

    request, response = application.make("request"), application.make("response")

    start_response(
        response.get_status_code(),
        response.get_headers() + response.cookie_jar.render_response(),
    )

    """Final Step
    This will take the data variable from the Service Container and return
    it to the WSGI server.
    """
    return (request, response)



from ..response import Response


class RedirectController:
    def __init__(self, url, status):
        self.url = url
        self.status = status

    def redirect(self, response: Response):
        return response.redirect(self.url, status=self.status)



"""The Masonite Response Object."""

import json
import mimetypes
from pathlib import Path
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ..foundation import Application

from ..routes.Router import Router
from ..exceptions import ResponseError, InvalidHTTPStatusCode, RouteNotFoundException
from ..headers import HeaderBag, Header
from ..utils.http import HTTP_STATUS_CODES
from ..utils.str import add_query_params
from ..cookies import CookieJar


class Response:
    """A Response object to be used to abstract the logic of getting a response ready to be
    returned."""

    def __init__(self, app: "Application"):
        self.app = app
        self.content: str = ""
        self._status: str = None
        self.statuses: dict = HTTP_STATUS_CODES
        self.header_bag = HeaderBag()
        self.cookie_jar = CookieJar()
        self.original = None

    def json(self, payload: Any, status: int = 200) -> bytes:
        """Set the response as a JSON response."""
        self.content = bytes(json.dumps(payload, ensure_ascii=False), "utf-8")
        self.make_headers(content_type="application/json; charset=utf-8")
        self.status(status)

        return self.data()

    def make_headers(self, content_type: str = "text/html; charset=utf-8") -> None:
        """Recompute Content-Length of the response after modyifing it."""
        self.header_bag.add(Header("Content-Length", str(len(self.to_bytes()))))

        # If the user did not change it directly
        self.header_bag.add_if_not_exists(Header("Content-Type", content_type))

    def header(self, name: str, value: str = None) -> "None|str":
        """Get response header for the given name if no value provided or add headers to response."""
        if value is None and isinstance(name, dict):
            for name, value in name.items():
                self.header_bag.add(Header(name, str(value)))
        elif value is None:
            header = self.header_bag.get(name)
            if isinstance(header, str):
                return header
            return header.value

        return self.header_bag.add(Header(name, value))

    def with_headers(self, headers: dict) -> "Response":
        """Add headers dictionary to response headers."""
        for name, value in headers.items():
            self.header_bag.add(Header(name, str(value)))
        return self

    def get_headers(self) -> list:
        return self.header_bag.render()

    def cookie(self, name: str, value: str = None, **options) -> "None|str":
        """Get response cookie for the given name if no value provided or add cookie to
        the response with the given name, value and options."""
        if value is None:
            cookie = self.cookie_jar.get(name)
            if not cookie:
                return
            return cookie.value

        return self.cookie_jar.add(name, value, **options)

    def delete_cookie(self, name: str) -> "Response":
        """Delete the cookie with the given name from the response."""
        self.cookie_jar.delete(name)
        return self

    def get_response_content(self) -> bytes:
        """Get response content."""
        return self.data()

    def status(self, status: "str|int") -> "Response":
        """Set HTTP status code of the response."""
        if isinstance(status, str):
            self._status = status
        elif isinstance(status, int):
            try:
                self._status = self.statuses[status]
            except KeyError:
                raise InvalidHTTPStatusCode
        return self

    def is_status(self, code: int) -> bool:
        """Check if response has the given status code."""
        return self._get_status_code_by_value(self.get_status_code()) == code

    def _get_status_code_by_value(self, value: int) -> "str|None":
        for key, status in self.statuses.items():
            if status == value:
                return key

        return None

    def get_status_code(self) -> str:
        """Gets the HTTP status code of the response as a human string, like "200 OK"."""
        return self._status

    def get_status(self):
        return self._get_status_code_by_value(self.get_status_code())

    def data(self) -> bytes:
        """Get the response content as bytes."""
        if isinstance(self.content, str):
            return bytes(self.content, "utf-8")

        return self.content

    def converted_data(self) -> "str|bytes":
        """Get the response content as string or bytes so that the WSGI server handles it."""
        if isinstance(self.data(), (dict, list)):
            return json.dumps(self.data(), ensure_ascii=False)
        else:
            return self.data()

    def view(self, view: Any, status: int = 200) -> "bytes|Response":
        """Set the response as a string or view."""
        self.original = view

        if isinstance(view, tuple):
            view, status = view
            self.status(status)

        if not self.get_status_code():
            self.status(status)

        if isinstance(view, (dict, list)):
            return self.json(view, status=self.get_status_code())
        elif hasattr(view, "serialize"):
            return self.json(view.serialize(), status=self.get_status_code())
        elif isinstance(view, int):
            view = str(view)
        elif hasattr(view, "get_response"):
            view = view.get_response()
        elif view is None:
            raise ResponseError(
                "Responses cannot be of type: None. Did you return anything in your responsable method?"
            )

        self.content = view

        self.make_headers()

        return self

    def back(self) -> "Response":
        """Set the response as a redirect response back to previous path defined from the
        request."""
        return self.redirect(url=self.app.make("request").get_back_path())

    def redirect(
        self,
        location: str = None,
        name: str = None,
        params: dict = {},
        url: str = None,
        status: int = 302,
        query_params={},
    ) -> "Response":
        """Set the response as a redirect response. The redirection location can be defined
        with the location URL or with a route name. If a route name is used, route params can
        be provided."""

        self.status(status)

        if location:
            location = add_query_params(location, query_params)
            self.header_bag.add(Header("Location", location))
        elif name:
            url = self._get_url_from_route_name(name, params, query_params)
            self.header_bag.add(Header("Location", url))
        elif url:
            url = add_query_params(url, query_params)
            self.header_bag.add(Header("Location", url))
        self.view("Redirecting ...")
        return self

    def _get_url_from_route_name(
        self, name: str, params: dict = {}, query_params: dict = {}
    ) -> str:
        route = self.app.make("router").find_by_name(name)
        if not route:
            raise RouteNotFoundException(f"Could not find route with the name '{name}'")
        return Router.compile_to_url(route.url, params, query_params)

    def to_bytes(self) -> "bytes":
        """Converts the response to bytes."""
        return self.converted_data()

    def download(self, name: str, location: str, force: bool = False) -> "Response":
        """Set the response as a file download response."""
        if force:
            self.header("Content-Type", "application/octet-stream")
            self.header(
                "Content-Disposition",
                'attachment; filename="{}{}"'.format(name, Path(location).suffix),
            )
        else:
            self.header("Content-Type", mimetypes.guess_type(location)[0])

        with open(location, "rb") as filelike:
            data = filelike.read()

        return self.view(data)



import os

from .PublishableResource import PublishableResource


class Package:
    def __init__(self):
        # absolute root directory to python package root
        self.abs_root = ""
        # relative import path to python package root
        self.module_root = ""
        self.name = ""
        self.config = ""
        self.commands = []
        self.views = []
        self.migrations = []
        self.controller_locations = []
        self.routes = []
        self.assets = []
        # all files to be published (can be assets, config, views...)
        self.resources = {}

    def _build_path(self, rel_path):
        """Build absolute path to package file."""
        return os.path.join(self.abs_root, rel_path)

    def _build_module_path(self, rel_path):
        """Build relative path to package file."""
        return os.path.join(self.module_root, rel_path)

    def add_config(self, config_path):
        self.config = self._build_module_path(config_path)
        return self

    def add_views(self, location):
        views_folder = self._build_module_path(location)
        if not os.path.isdir(self._build_path(location)):
            raise ValueError(
                "views() first argument must be a folder containing all package views."
            )
        self.views = views_folder
        return self

    def add_migrations(self, *migrations):
        for migration in migrations:
            self.migrations.append(self._build_module_path(migration))
        return self

    def add_routes(self, *routes):
        for route in routes:
            self.routes.append(self._build_module_path(route))
        return self

    def add_assets(self, *assets):
        for asset in assets:
            self.assets.append(self._build_module_path(asset))
        return self

    def add_controller_locations(self, *controller_locations):
        for loc in controller_locations:
            self.controller_locations.append(self._build_module_path(loc))
        return self

    def add_publishable_resource(self, key, source, abs_destination):
        resource = self.resources.get(key, None) or PublishableResource(key)
        abs_path = self._build_path(source)
        resource.add(abs_path, abs_destination)
        self.resources.update({key: resource})



import os
from cleo import Application as CommandApplication
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .Application import Application

from .response_handler import response_handler
from .. import __version__
from ..commands import (
    TinkerCommand,
    CommandCapsule,
    KeyCommand,
    ServeCommand,
    QueueWorkCommand,
    QueueRetryCommand,
    QueueTableCommand,
    QueueFailedCommand,
    AuthCommand,
    MakePolicyCommand,
    MakeControllerCommand,
    MakeJobCommand,
    MakeMailableCommand,
    MakeProviderCommand,
    PublishPackageCommand,
    MakeTestCommand,
    DownCommand,
    UpCommand,
    MakeCommandCommand,
    MakeViewCommand,
    MakeMiddlewareCommand,
    PresetCommand,
)
from ..environment import LoadEnvironment
from ..middleware import MiddlewareCapsule
from ..routes import Router
from ..loader import Loader

from ..tests.HttpTestResponse import HttpTestResponse
from ..tests.TestResponseCapsule import TestResponseCapsule


class Kernel:
    def __init__(self, app: "Application"):
        self.application = app

    def register(self) -> None:
        """Register core Masonite features in the project."""
        self.load_environment()
        self.register_framework()
        self.register_commands()
        self.register_testing()

    def load_environment(self) -> None:
        """Load environment variables into the application."""
        LoadEnvironment()

    def register_framework(self) -> None:
        self.application.set_response_handler(response_handler)
        self.application.use_storage_path(
            os.path.join(self.application.base_path, "storage")
        )
        self.application.bind("middleware", MiddlewareCapsule())
        self.application.bind(
            "router",
            Router(),
        )
        self.application.bind("loader", Loader())

    def register_commands(self) -> None:
        self.application.bind(
            "commands",
            CommandCapsule(CommandApplication("Masonite", __version__)).add(
                TinkerCommand(),
                KeyCommand(),
                ServeCommand(self.application),
                QueueWorkCommand(self.application),
                QueueRetryCommand(self.application),
                QueueFailedCommand(),
                QueueTableCommand(),
                AuthCommand(self.application),
                MakePolicyCommand(self.application),
                MakeControllerCommand(self.application),
                MakeJobCommand(self.application),
                MakeMailableCommand(self.application),
                MakeProviderCommand(self.application),
                PublishPackageCommand(self.application),
                MakeTestCommand(self.application),
                DownCommand(),
                UpCommand(),
                MakeCommandCommand(self.application),
                MakeViewCommand(self.application),
                MakeMiddlewareCommand(self.application),
                PresetCommand(self.application),
            ),
        )

    def register_testing(self) -> None:
        test_response = TestResponseCapsule(HttpTestResponse)
        self.application.bind("tests.response", test_response)



from .Limit import Limit
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..request import Request


class Limiter:
    def __init__(self):
        pass

    def allow(self, request: "Request"):
        pass


class GlobalLimiter(Limiter):
    """Apply a global limit"""

    def __init__(self, limit: str):
        self.limit = limit

    def allow(self, request: "Request") -> "Limit":
        return Limit.from_str(self.limit)


class UnlimitedLimiter(Limiter):
    def allow(self, request: "Request") -> "Limit":
        return Limit.unlimited()


class GuestsOnlyLimiter(Limiter):
    """Apply a limit for guests only."""

    def __init__(self, limit: str):
        self.limit = limit

    def allow(self, request: "Request") -> "Limit":
        if request.user():
            return Limit.unlimited()
        else:
            return Limit.from_str(self.limit).by(request.ip())



"""Helpers for working with HTTP."""


HTTP_STATUS_CODES = {
    100: "100 Continue",
    101: "101 Switching Protocol",
    102: "102 Processing",
    103: "Early Hints",
    200: "200 OK",
    201: "201 Created",
    202: "202 Accepted",
    203: "203 Non-Authoritative Information",
    204: "204 No Content",
    205: "205 Reset Content",
    206: "206 Partial Content",
    207: "207 Multi-Status",
    208: "208 Multi-Status",
    226: "226 IM Used",
    300: "300 Multiple Choice",
    301: "301 Moved Permanently",
    302: "302 Found",
    303: "303 See Other",
    304: "304 Not Modified",
    307: "307 Temporary Redirect",
    308: "308 Permanent Redirect",
    400: "400 Bad Request",
    401: "401 Unauthorized",
    402: "402 Payment Required",
    403: "403 Forbidden",
    404: "404 Not Found",
    405: "405 Method Not Allowed",
    406: "406 Not Acceptable",
    407: "407 Proxy Authentication Required",
    408: "408 Request Timeout",
    409: "409 Conflict",
    410: "410 Gone",
    411: "411 Length Required",
    412: "412 Precondition Failed",
    413: "413 Payload Too Large",
    414: "414 URI Too Long",
    415: "415 Unsupported Media Type",
    416: "416 Requested Range Not Satisfiable",
    417: "417 Expectation Failed",
    418: "418 I'm a teapot",
    421: "421 Misdirected Request",
    422: "422 Unprocessable Entity",
    423: "423 Locked",
    424: "424 Failed Dependency",
    425: "425 Too Early",
    426: "426 Upgrade Required",
    428: "428 Precondition Required",
    429: "429 Too Many Requests",
    431: "431 Request Header Fields Too Large",
    451: "451 Unavailable For Legal Reasons",
    500: "500 Internal Server Error",
    501: "501 Not Implemented",
    502: "502 Bad Gateway",
    503: "503 Service Unavailable",
    504: "504 Gateway Timeout",
    505: "505 HTTP Version Not Supported",
    506: "506 Variant Also Negotiates",
    507: "507 Insufficient Storage",
    508: "508 Loop Detected",
    510: "510 Not Extended",
    511: "511 Network Authentication Required",
}


def generate_wsgi(wsgi={}, path="/", query_string="", method="GET"):
    """Generate the WSGI environment dictionary that we receive from a HTTP request."""
    import io

    data = {
        "wsgi.version": (1, 0),
        "wsgi.multithread": False,
        "wsgi.multiprocess": True,
        "wsgi.run_once": False,
        "wsgi.input": io.BytesIO(),
        "SERVER_SOFTWARE": "gunicorn/19.7.1",
        "REQUEST_METHOD": method,
        "QUERY_STRING": query_string,
        "RAW_URI": path,
        "SERVER_PROTOCOL": "HTTP/1.1",
        "HTTP_HOST": "127.0.0.1:8000",
        "HTTP_ACCEPT": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
        "HTTP_UPGRADE_INSECURE_REQUESTS": "1",
        "HTTP_COOKIE": "",
        "HTTP_USER_AGENT": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_2) AppleWebKit/604.4.7 (KHTML, like Gecko) Version/11.0.2 Safari/604.4.7",
        "HTTP_ACCEPT_LANGUAGE": "en-us",
        "HTTP_ACCEPT_ENCODING": "gzip, deflate",
        "HTTP_CONNECTION": "keep-alive",
        "wsgi.url_scheme": "http",
        "REMOTE_ADDR": "127.0.0.1",
        "REMOTE_PORT": "62241",
        "SERVER_NAME": "127.0.0.1",
        "SERVER_PORT": "8000",
        "PATH_INFO": path,
        "SCRIPT_NAME": "",
    }
    data.update(wsgi)
    return data
